"""Utilities for working with common plots."""

from collections.abc import Callable, Mapping
from functools import reduce, wraps
from pathlib import Path
from typing import Any, Concatenate, Literal, ParamSpec, TextIO, TypeAlias, cast

import pandas as pd
import plotly.graph_objects as go

ImageFormat: TypeAlias = Literal["svg", "pdf", "png", "jpg", "webp"]


def export_html(
    fig: go.Figure,
    out: Path | str | TextIO,
    width: int = 800,
    height: int = 450,
    scale: float = 3,
    name: str = "plot",
    download_format: ImageFormat = "svg",
):
    """Save plotly figure to interactive html.

    Args:
        fig: Plotly figure to save.
        out: Path to save to or file-like object to write to.
        width: Width of the figure in pixels.
        height: Height of the figure in pixels.
        scale: Scale factor for the figure.
        name: Name of the figure.
        download_format: Format to use when downloading the figure.
    """
    fig.write_html(
        out,
        include_plotlyjs="cdn",
        full_html=False,
        config={
            "responsive": True,
            "toImageButtonOptions": {
                "format": download_format,
                "filename": name,
                "height": height,
                "width": width,
                "scale": scale,
            },
        },
    )


P = ParamSpec("P")
PlottingFunction = Callable[Concatenate[pd.DataFrame, P], go.Figure]


def _merge_data(figs: Mapping[str, go.Figure | go.Frame]) -> list:
    return [d for fig in figs.values() for d in (fig.data or [])]


def _merge_layout(figs: Mapping[str, go.Figure | go.Frame]) -> dict[str, Any]:
    return reduce(
        lambda a, b: {**a, **b},
        [
            cast(dict, fig.layout.to_plotly_json())  # type: ignore
            for fig in figs.values()
        ],
    )


def _get_frames(figs: Mapping[str, go.Figure], label: str) -> dict[str, go.Frame]:
    return {
        group: frame
        for group, fig in figs.items()
        for f in fig.frames
        if (frame := cast(go.Frame, f)).name == label
    }


go.Frame


def with_dropdown(
    group_by: str,
    dropdown_kwars: dict[str, Any] | None = None,
) -> Callable[[PlottingFunction[P]], PlottingFunction[P]]:
    """Add a dropdown to a plotting function.

    Args:
        group_by: Column to group by for the dropdown.
        dropdown_kwars: Keyword arguments to pass to the dropdown layout.

    Returns:
        Decorator that adds a dropdown to the returned plot.
    """

    def decorator(func: PlottingFunction[P]) -> PlottingFunction[P]:
        """Add a dropdown to a plotting function."""

        @wraps(func)
        def wrapper(data: pd.DataFrame, *args: P.args, **kwargs: P.kwargs) -> go.Figure:
            figs = {
                str(group): func(sub_data, *args, **kwargs)
                for group, sub_data in data.groupby(group_by)
            }

            fig_data = _merge_data(figs)
            layout = _merge_layout(figs)

            frames = None
            if "sliders" in layout:
                labels = [step["label"] for step in layout["sliders"][0]["steps"]]
                frames = []
                for label in labels:
                    sub_frames = _get_frames(figs, label)
                    frames.append(
                        go.Frame(
                            data=_merge_data(sub_frames),
                            layout=_merge_layout(sub_frames),
                            name=label,
                        )
                    )

            split_fig = go.Figure(
                data=fig_data,
                layout=layout,
                frames=frames,
            )

            trace_sources = pd.Series(
                [group for group, fig in figs.items() for _ in fig.data]
            )

            split_fig.update_layout(
                updatemenus=[
                    *(
                        [split_fig.layout["updatemenus"][0]]  # type: ignore
                        if "updatemenus" in layout
                        else []
                    ),
                    {
                        "active": 0,
                        "buttons": [
                            {
                                "label": "All",
                                "method": "update",
                                "args": [
                                    {
                                        "visible": [True] * len(trace_sources),
                                    },
                                ],
                            },
                            *(
                                {
                                    "label": group,
                                    "method": "update",
                                    "args": [
                                        {
                                            "visible": (
                                                trace_sources == str(group)
                                            ).to_list(),
                                        },
                                    ],
                                }
                                for group in figs.keys()
                            ),
                        ],
                        "direction": "up",
                        "pad": {"r": 10, "t": 70},
                        "showactive": True,
                        "x": 1.1,
                        "xanchor": "left",
                        "y": 0,
                        "yanchor": "top",
                        **(dropdown_kwars or {}),
                    },
                ]
            )

            return split_fig

        return wrapper

    return decorator



"""Test stats module."""

import pandas as pd
from py_research.stats import dist_table


def test_dist_table():
    """Test dist_table."""
    # Test with a DataFrame
    df = pd.DataFrame(
        {
            "id": ["A", "A", "B", "B", "B", "C"],
            "category": ["X", "Y", "X", "Y", "X", "Y"],
            "value": [1, 2, 3, 4, 5, 6],
        }
    )
    result = dist_table(df, "id", "category")
    expected = pd.Series(
        [2, 3], index=pd.Index(["X", "Y"], name="category"), name="freq"
    )
    pd.testing.assert_series_equal(result, expected)

    # Test with multiple id_cols and category_cols
    df["sub_id"] = ["P", "Q", "P", "Q", "Q", "Q"]
    df["sub_category"] = ["U", "V", "U", "V", "U", "V"]
    result = dist_table(df, ["id", "sub_id"], ["category", "sub_category"])
    expected = pd.Series(
        [3, 3],
        index=pd.MultiIndex.from_tuples(
            [
                ("X", "U"),
                ("Y", "V"),
            ],
            names=["category", "sub_category"],
        ),
        name="freq",
    )
    pd.testing.assert_series_equal(result, expected)

    # Test with None id_cols
    df = df.set_index(["id", "sub_id"])
    result = dist_table(df, None, ["category", "sub_category"])
    pd.testing.assert_series_equal(result, expected)



"""Utils for Python code reflection."""

import inspect
import platform
import re
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from functools import cache, reduce
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Any, Generic, TypeVar
from urllib.parse import urlparse

import importlib_metadata as meta
import numpy as np
import requests
from git import Repo
from packaging.requirements import Requirement
from packaging.specifiers import Specifier
from packaging.version import InvalidVersion, Version

T = TypeVar("T")
T2 = TypeVar("T2")


def _get_calling_frame(offset=0):
    stack = inspect.stack()
    if len(stack) < offset + 3:
        raise RuntimeError("No caller!")
    return stack[offset + 2]


def get_calling_module_name() -> str | None:
    """Return the name of the module calling the current function."""
    mod = inspect.getmodule(_get_calling_frame(offset=1).frame)
    return mod.__name__ if mod is not None else None


def get_full_args_dict(
    func: Callable, args: Sequence, kwargs: dict[str, Any] | None = None
) -> dict[str, Any]:
    """Return dict of all args + kwargs.

    Args:
        func: Function to inspect.
        args: Positional arguments given to the function.
        kwargs: Keyword arguments given to the function.

    Returns:
        Dictionary of all args + kwargs.
    """
    argspec = inspect.getfullargspec(func)

    arg_defaults = argspec.defaults or []
    kwdefaults = dict(zip(argspec.args[-len(arg_defaults) :], arg_defaults))

    posargs = dict(zip(argspec.args[: len(args)], args))

    return {**kwdefaults, **posargs, **(kwargs or {})}


def get_return_type(func: Callable) -> type | None:
    """Get the return type annotation of given function, if any."""
    sig = inspect.signature(func)
    return (
        sig.return_annotation
        if sig.return_annotation != inspect.Signature.empty
        else None
    )


def get_all_subclasses(cls: type[T]) -> set[type[T]]:
    """Return all subclasses of given class."""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)]
    )


@cache
def _get_distributions() -> dict[str, meta.Distribution]:
    return {d.metadata["Name"]: d for d in meta.distributions()}


def _get_module_file(module: ModuleType) -> Path | None:
    """Get the file of given module, if any."""
    return (
        Path(str(module.__file__))
        if hasattr(module, "__file__") and module.__file__ is not None
        else None
    )


def _file_url_to_path(file_url: str):
    # Get the operating system name
    os_name = platform.system()

    # Parse the file URL
    parsed_url = urlparse(file_url)

    # Extract the path component from the parsed URL
    file_path = parsed_url.path

    # Convert the path to a Windows-compatible path
    # On Windows, you may need to remove the leading '/' from the path
    if os_name == "Windows" and file_path.startswith("/"):
        file_path = file_path[1:]

    # Create a Path object
    path_object = Path(file_path)

    return path_object


def get_module_distribution(module: ModuleType) -> meta.Distribution | None:
    """Get the distribution package of given module, if any."""
    mod_file = _get_module_file(module)
    if mod_file is None:
        return None

    dists = {
        Path(dist.locate_file(f"{name}")): dist
        for name, dist in _get_distributions().items()
    }

    mod_dists = []
    for dist_path, dist in dists.items():
        if mod_file.is_relative_to(dist_path):
            mod_dists.append(dist)
        elif dist.origin is not None and "url" in dist.origin.__dict__:  # type: ignore
            url = str(dist.origin.url)
            if url.startswith("file://"):
                path = _file_url_to_path(url)
                if mod_file.is_relative_to(path):
                    mod_dists.append(dist)

    return mod_dists[0] if len(mod_dists) > 0 else None


def get_all_module_dependencies(
    module: ModuleType,
    _ext_deps: set[str] | None = None,
    _int_deps: set[ModuleType] | None = None,
) -> tuple[set[str], set[ModuleType]]:
    """Return all (indirect) dependency modules of given module.

    Args:
        module: Module to inspect.

    Returns:
        Tuple of external and internal dependencies.
    """
    if _ext_deps is None:
        _ext_deps = set()
    if _int_deps is None:
        _int_deps = set()

    deps = [
        dep
        for _, m in inspect.getmembers(module)
        if (dep := inspect.getmodule(m)) is not None
    ]
    ext_deps_map = {
        dep: dist.metadata["Name"]
        for dep in deps
        if (dist := get_module_distribution(dep)) is not None
    }
    new_ext_deps = set(ext_deps_map.values()) - _ext_deps

    these_int_deps = set(deps) - set(ext_deps_map.keys())
    new_int_deps = these_int_deps - _int_deps

    if len(new_ext_deps) > 0 or len(new_int_deps) > 0:
        _ext_deps |= new_ext_deps
        _int_deps |= new_int_deps

        sub_ext_deps, sub_int_deps = zip(
            *[
                get_all_module_dependencies(d, _ext_deps, _int_deps)
                for d in new_int_deps
            ]
        )

        return set.union(*sub_ext_deps), set.union(*sub_int_deps)
    else:
        return _ext_deps, _int_deps


def get_module_repo(module: ModuleType) -> Repo | None:
    """Get the Git repository of given module, if any."""
    mod_file = _get_module_file(module)
    if mod_file is None:
        return None

    return Repo(mod_file, search_parent_directories=True)


def _check_version(lower_bound: str, version: str) -> bool:
    """Check if a version is compatible with a lower bound."""
    lb = tuple(map(int, lower_bound.split(".")))
    v = tuple(map(int, version.split(".")))
    return v >= lb and v[0] == lb[0]


@dataclass
class PyObjectRef(Generic[T]):
    """Reference to a static Python object."""

    type: type[T]
    """Type of the object."""

    package: str
    """Name of the package the object belongs to."""

    module: str
    """Name of the module the object belongs to."""

    object: str
    """Name of the object."""

    version: str | None = None
    """Version of the package or object."""

    repo: str | None = None
    """URL of the package's repository."""

    revision: str | None = None
    """Revision of the repo."""

    @staticmethod
    def _follows_semver(version: str) -> bool:
        return re.match(r"^\d+\.\d+\.\d+$", version) is not None

    @staticmethod
    def reference(obj: T2, version: str | None = None) -> "PyObjectRef[T2]":
        """Create a reference to given object."""
        qualname = getattr(obj, "__qualname__")
        if qualname is None:
            raise ValueError("Object must have fully qualified name (`__qualname__`)")

        module = inspect.getmodule(obj)
        if module is None:
            raise ValueError("Object must be associated with a module.")

        dist = get_module_distribution(module)
        if dist is None:
            raise ValueError("Object must be associated with a package.")

        url: str | None = (
            dict(dist.origin.__dict__).get("url") if dist.origin is not None else None
        )

        repo = None
        if url is not None and str(url).startswith("file://"):
            repo = get_module_repo(module)
            if repo is not None:
                url = repo.remote().url

        return PyObjectRef(
            type=type(obj),
            package=dist.name,
            module=module.__name__,
            object=qualname,
            version=(
                version
                if version
                else getattr(obj, "__version__")
                if hasattr(obj, "__version__")
                else dist.version
                if PyObjectRef._follows_semver(dist.version)
                else None
            ),
            repo=(url if url else "https://pypi.org"),
            revision=repo.head.commit.hexsha if repo is not None else None,
        )

    def resolve(self) -> T:
        """Resolve object reference."""
        dist = _get_distributions().get(self.package)
        if dist is None or (
            self.version is not None and not _check_version(self.version, dist.version)
        ):
            raise ImportError(
                f"Please install package '{self.package}' with version '{self.version}'"
                f"from '{self.repo}' to resolve this object reference."
            )

        try:
            module = import_module(self.module)
        except ModuleNotFoundError:
            raise ImportError(
                f"Cannot import module '{self.module}' from package '{self.package}'."
            )

        url: str | None = (
            dict(dist.origin.__dict__).get("url") if dist.origin is not None else None
        )
        if url is not None:
            if str(url).startswith("file://"):
                repo = get_module_repo(module)
                if repo is not None:
                    url = repo.remote().url
            if url != self.repo:
                raise ImportError(
                    f"URL mismatch: Package '{self.package} should be from "
                    f"'{self.repo}' but is from '{url}'."
                )

        obj = reduce(getattr, self.object.split("."), module)
        if not isinstance(obj, self.type):
            raise TypeError(
                f"Object `{'.'.join([self.package, self.module, self.object])}` "
                f"must have type `{self.type}`"
            )

        if (
            self.version is not None
            and hasattr(obj, "__version__")
            and not _check_version(
                self.version,
                given_version := getattr(obj, "__version__"),
            )
        ):
            raise ValueError(
                f"Requested version {self.version} is not compatible with "
                + f"existing version {given_version}."
            )

        return obj


def stref(obj: Any) -> str:
    """Get string representation of given object reference."""
    obj_ref = PyObjectRef.reference(obj)
    return f"{obj_ref.module}.{obj_ref.object}"


def get_dist_requirements(dist: meta.Distribution) -> list[Requirement] | None:
    """Get a list of declared packages via pdm."""
    return (
        [Requirement(dep) for dep in dist.requires]
        if dist.requires is not None
        else None
    )


def get_versions_on_pypi(package: meta.Distribution | str) -> set[Version]:
    """Get all available versions of given distribution."""
    if isinstance(package, meta.Distribution):
        if package.origin is not None:
            return set()
        package = package.name

    url = f"https://pypi.org/pypi/{package}/json"

    response = requests.get(url)
    if response.status_code == 404:
        return set()

    data = response.json()

    versions = set()
    for v in data["releases"].keys():
        try:
            versions.add(Version(v))
        except InvalidVersion:
            pass

    return versions


def version_diff(v1: Version, v2: Version) -> Version | None:
    """Get the difference between two versions (v1 - v2).

    if v1 is smaller than v2, returns None.
    """
    if v1 < v2:
        return None

    v1_arr = np.array((v1.major, v1.minor, v1.micro))
    v2_arr = np.array((v2.major, v2.minor, v2.micro))

    diff = v1_arr - v2_arr

    if diff[0] > 0:
        diff[1:] = v1_arr[1:]
    elif diff[1] > 0:
        diff[2] = v1_arr[2]

    return Version(".".join(str(v) for v in diff))


def get_outdated_deps(
    dist: meta.Distribution,
    allowed_diff: Specifier = Specifier("<=1.1.1"),
) -> dict[str, tuple[Version, Version]]:
    """Get a list of outdated dependencies of a distribution.

    Args:
        dist: Distribution to inspect.
        allowed_diff: Allowed difference between current and latest version.

    Returns:
        Dictionary of outdated package names with current and latest version.
    """
    deps = get_dist_requirements(dist)

    if deps is None:
        return {}

    outdated = {}
    for dep in deps:
        try:
            dep_dist = meta.distribution(dep.name)
        except meta.PackageNotFoundError:
            dep_dist = dep.name

        versions = get_versions_on_pypi(dep_dist)
        versions = {
            v
            for v in versions
            if not v.is_prerelease and not v.is_postrelease and not v.is_devrelease
        }

        matching_req = set(dep.specifier.filter(versions))
        newest_matching = max(matching_req)

        newer = versions - matching_req
        if len(newer) == 0:
            continue

        newest = max(newer)
        diff = version_diff(newest, newest_matching)
        if diff is not None and diff not in allowed_diff:
            outdated[dep.name] = (newest_matching, newest)

    return outdated



"""Test the caching module."""

import datetime
from pathlib import Path
from tempfile import gettempdir

from py_research.caching import FileCache, get_cache


def test_get_cache_custom_params():
    """Test the custom parameters of get_cache."""
    custom_root_path = Path(gettempdir())
    custom_max_cache_time = datetime.timedelta(days=1)

    # Create via positional args.
    cache1 = get_cache("custom_cache", custom_root_path, custom_max_cache_time)

    # Assert positional args result in correct instance.
    assert isinstance(cache1, FileCache)
    assert (
        cache1.path
        == custom_root_path / "tests.py_research.test_caching" / "custom_cache"
    )
    assert cache1.max_cache_time == custom_max_cache_time

    # Create via keyword args.
    cache2 = get_cache(
        name="custom_cache",
        root_path=custom_root_path,
        max_cache_time=custom_max_cache_time,
    )

    # Assert keyword args result in correct instance.
    assert isinstance(cache2, FileCache)
    assert (
        cache2.path
        == custom_root_path / "tests.py_research.test_caching" / "custom_cache"
    )
    assert cache2.max_cache_time == custom_max_cache_time


def test_cache_function():
    """Test the cache_function decorator."""
    tempdir = Path(gettempdir())

    # Create cache.
    cache = get_cache("test_cache", tempdir)

    # Assert cache dir was created.
    assert cache.path == tempdir / "tests.py_research.test_caching" / "test_cache"
    assert (tempdir / "tests.py_research.test_caching" / "test_cache").exists()

    # Create function to cache.
    @cache.function(id_arg_subset=["a", "b"])
    def test_func(a: int, b: int, c: int) -> int:
        return a + b + c

    # Assert function returns correct value.
    assert test_func(1, 2, 3) == 6

    # Assert function returns cached value for same id args.
    assert test_func(1, 2, 4) == 6

    # Assert function returns new value for different id args.
    assert test_func(1, 3, 4) == 8

    # Create another function to cache.
    @cache.function(id_arg_subset=["a"], use_raw_arg=True)
    def test_func2(a: int, b: int) -> int:
        return a + b

    # Assert function returns correct value.
    assert test_func2(1, 2) == 3

    # Assert function returns cached value for same id args.
    assert test_func2(1, 3) == 3

    # Assert function returns new value for different id args.
    assert test_func2(2, 3) == 5



"""Utilities for producing hashes of common objects."""

import hashlib
from collections.abc import Sequence
from datetime import date, datetime, time, timedelta
from functools import reduce
from numbers import Number
from typing import Any

import pandas as pd
from pandas.util import hash_pandas_object


def _stable_hash(
    data: Number | str | bytes | date | time | datetime | timedelta,
) -> int:
    m = hashlib.md5(usedforsecurity=False)
    m.update(bytes(str(data), "utf-8"))
    return int.from_bytes(m.digest())


def _hash_sequence(s: Sequence) -> int:
    return reduce(lambda x, y: gen_int_hash(str(x) + str(gen_int_hash(y))), s, 0)


def gen_int_hash(obj: Any) -> int:
    """Generate stable hash for obj (must be known, hashable or composed of such)."""
    match obj:
        case (Number() | str() | bytes() | date() | time() | datetime() | timedelta()):
            return _stable_hash(obj)
        case pd.DataFrame() | pd.Series() | pd.Index():
            return sum(hash_pandas_object(obj))
        case list() | tuple():
            return _hash_sequence(obj)
        case dict():
            return sum(_hash_sequence(item) for item in obj.items())
        case set():
            return sum(gen_int_hash(item) for item in obj)
        case None:
            return 0
        case _:
            if hasattr(obj, "__getstate__"):
                state = obj.__getstate__()
                state = state if isinstance(state, tuple) else (state,)
                state_dicts = [s for s in state if isinstance(s, dict)]

                if len(state_dicts) > 0:
                    return sum(
                        sum(_hash_sequence(item) for item in s.items())
                        for s in state_dicts
                    )
            if hasattr(obj, "__getnewargs_ex__"):
                args, kwargs = obj.__getnewargs_ex__()
                return sum(gen_int_hash(item) for item in args) + sum(
                    gen_int_hash(item) for item in kwargs.items()
                )
            if hasattr(obj, "__getnewargs__"):
                args = obj.__getnewargs__()
                return sum(gen_int_hash(item) for item in args)

    raise ValueError()


def gen_str_hash(x: Any, length: int = 10) -> str:
    """Generate stable hash for obj (must be known, hashable or composed of such).

    Args:
        x: Object to hash.
        length: Length of the hash.
        raw_str:
            Whether to use the raw string representation of the object,
            if it is a string.

    Returns:
        Hash of the object as string.
    """
    s = str(abs(gen_int_hash(x)))
    return s[:length].rjust(length, "0")



"""Simple & semi-automatic intl. + localization of data analysis functions."""

from collections.abc import Callable, Generator, Iterable, Mapping
from contextvars import ContextVar
from dataclasses import asdict, dataclass, field
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from functools import partial
from locale import LC_ALL, getlocale, normalize, setlocale
from os import environ
from pathlib import Path
from typing import Any, Literal, ParamSpec, TypeVar, cast

import pandas as pd
from babel import Locale, UnknownLocaleError
from babel.dates import (
    format_date,
    format_datetime,
    format_interval,
    format_time,
    format_timedelta,
)
from babel.numbers import format_decimal
from deep_translator import GoogleTranslator
from pydantic import parse_obj_as
from pydantic.dataclasses import dataclass as pydantic_dataclass
from typing_extensions import Self
from yaml import CLoader, load

from py_research.caching import get_cache
from py_research.enums import StrEnum
from py_research.geo import Country, CountryScheme, GeoScheme
from py_research.hashing import gen_int_hash

cache = get_cache()

K = TypeVar("K")
V = TypeVar("V")
P = ParamSpec("P")


def _merge_ordered_dicts(d: list[dict[K, V]]) -> dict[K, V]:
    d = d if len(d) > 0 else [{}]
    merged = d[0].copy()

    for di in d[1:]:
        for k, v in di.items():
            merged[k] = v

    return merged


class DtUnit(StrEnum):
    """Common datetime units with their LDML pattern strings."""

    year = "yyyy"
    quarter = "QQQ"
    month = "MM"
    week = "ww"
    week_of_month = "W"
    day = "dd"
    day_of_year = "DD"
    hour = "HH"
    minute = "mm"
    second = "ss"
    millisecond = "SSS"


@pydantic_dataclass
class Format:
    """Custom options for localized formatting."""

    decimal_notation: Literal["plain", "scientific", "percent"] | None = None
    """How to format decimals."""

    decimal_digits: int | None = None
    """How many digits to show after the decimal point."""

    decimal_min_digits: int | None = None
    """Minimum number of digits to show after the decimal point."""

    decimal_max_digits: int | None = None
    """Maximum number of digits to show after the decimal point."""

    decimal_group_separator: bool | None = None
    """Whether to use decimal group separators."""

    datetime_auto_format: Literal["short", "medium", "long"] | None = None
    """How to auto-format datetimes."""

    datetime_format: DtUnit | str | None = None
    """`ldml pattern`_ for formatting datetime objects.
    .. _ldml pattern: https://www.unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table
    """

    timedelta_auto_format: Literal[
        "narrow", "short", "medium", "long", "raw"
    ] | None = None
    """How to auto-format timedeltas."""

    timedelta_resolution: Literal[
        "year", "month", "week", "day", "hour", "minute", "second"
    ] | None = None
    """Unit to use for formatting timedeltas. Has no effect for auto-format 'raw'."""

    country_format: CountryScheme | None = None
    """How to format countries."""

    fallback_to_translation: bool = True
    """Whether to fallback to text translation if no other formatting method matches."""

    na_representation: str = ""
    """String to use for NaN or None values."""

    def merge(self, other: Self) -> Self:
        """Merge and override ``self`` with ``other``."""
        return cast(
            Self,
            parse_obj_as(
                Format,
                {
                    **asdict(self),
                    **{k: v for k, v in asdict(other).items() if v is not None},
                },
            ),
        )

    def spec(self, t: type) -> str:
        """Generate format-spec matching this format usign current locale."""
        return get_localization().format_spec(t, self)

    def auto_digits(
        self, series: Iterable[float], min_sig: int = 3, fixed: bool = True
    ) -> Self:
        """Copy and change options according to numbers in ``sample``.

        Change such that all given numbers have minimum ``min_sig`` significant digits.

        Args:
            series: Series of numbers to use for determining options.
            min_sig: Minimum number of significant digits.
            fixed: Whether to fix the number of digits after the decimal point.

        Returns:
            New instance with changed options.
        """
        minimum = min(abs(i) for i in series)
        min_int_digits = (
            (
                len(str(i))
                if (
                    i := int(
                        minimum if self.decimal_notation == "plain" else minimum * 100
                    )
                )
                != 0
                else 0
            )
            if self.decimal_notation in ("plain", "percent")
            else 1
        )
        min_after_comma = min_sig - min_int_digits
        return cast(
            Self,
            Format(
                **{
                    **asdict(self),
                    **dict(
                        decimal_min_digits=min_after_comma,
                        decimal_max_digits=(min_after_comma if fixed else None),
                    ),
                }
            ),
        )


@dataclass
class Args:
    """Match for args of a text template."""

    values: Any = None
    all: bool = False

    def __hash__(self) -> int:  # noqa: D105
        return gen_int_hash(asdict(self))

    def match(self, args: tuple) -> float:
        """Check whether given args match this match."""
        if self.all:
            return 1

        values = self.values if isinstance(self.values, tuple | list) else [self.values]

        if len(args) != len(values):
            return 0

        matchers = [v if isinstance(v, set) else {v} for v in values]

        return max(
            (1 / len(matcher) if arg in matcher else 0)
            for arg, matcher in zip(args, matchers)
        )


@dataclass
class Template:
    """Override for a text template."""

    replace: str | None = None
    """Replace template with this template string."""

    wrap: str | None = None
    """Wrap template with this template string."""

    args: tuple | dict[int, Any] | None = None
    """Replace (some) args."""

    def apply_tmpl(self, tmpl: str) -> str | None:
        """Apply this template to given template string and args."""
        return (
            self.wrap.format(tmpl)
            if self.wrap is not None
            else self.replace
            if self.replace is not None
            else None
        )

    def apply_args(self, args: tuple) -> tuple | None:
        """Apply this template to given args."""
        return (
            self.args
            if isinstance(self.args, tuple)
            else tuple(
                self.args[i] if i in self.args else a for i, a in enumerate(args)
            )
            if isinstance(self.args, dict)
            else None
        )


@pydantic_dataclass
class Overrides:
    """Custom, language-dependent text overrides."""

    vocabulary: dict[str, str] = field(default_factory=dict)
    """Term overrides."""

    templates: Mapping[
        str, str | Template | dict[str | tuple | Args, str | Template]
    ] = field(default_factory=dict)
    """Template overrides by name."""

    format: Format = field(default_factory=Format)
    """Overrides for formatting."""

    def merge(self, other: Self) -> Self:
        """Merge and override ``self`` with ``other``."""
        texts = {}
        for name in set(self.templates.keys()).union(other.templates.keys()):
            self_texts = self.templates.get(name, {})
            other_texts = other.templates.get(name, {})

            texts[name] = _merge_ordered_dicts(
                [
                    self_texts
                    if isinstance(self_texts, dict)
                    else {
                        Args(all=True): Template(self_texts)
                        if isinstance(self_texts, str)
                        else self_texts
                    },
                    other_texts
                    if isinstance(other_texts, dict)
                    else {
                        Args(all=True): Template(other_texts)
                        if isinstance(other_texts, str)
                        else other_texts
                    },
                ]
            )

        return cast(
            Self,
            Overrides(
                vocabulary=_merge_ordered_dicts([self.vocabulary, other.vocabulary]),
                templates=texts,
                format=self.format.merge(other.format),
            ),
        )


texts_file_path = Path(environ.get("TEXTS_FILE_PATH") or (Path.cwd() / "texts.yaml"))

file_overrides = None
if texts_file_path.is_file():
    with open(texts_file_path, encoding="utf-8") as f:
        file_overrides = parse_obj_as(dict[str, Overrides], load(f, Loader=CLoader))


def _get_default_locale() -> Locale:
    """Get default locale, falling back to English."""
    try:
        return Locale.parse(normalize(getlocale()[0] or "en_US"), sep="_")
    except UnknownLocaleError:
        return Locale("en", "US")


_ldml_to_posix = {
    "a": "%p",  # AM or PM
    "dd": "%d",  # day of month, padded
    "d": "%e",  # day of month, unpadded
    "DDD": "%j",  # day of year, padded
    "DD": "%-j",  # day of year, unpadded
    "D": "%-j",  # day of year, unpadded
    "EEEE": "%A",  # weekday
    "hh": "%I",  # 12-hour time, padded
    "h": "%-I",  # 12-hour time, unpadded
    "HH": "%H",  # 24-hour time, padded
    "H": "%-H",  # 24-hour time, unpadded
    "MMMM": "%B",  # month: name, full
    "MMM": "%b",  # month name, short
    "MM": "%m",  # month, padded
    "M": "%-m",  # month, unpadded
    "mm": "%M",  # minute, padded
    "m": "%-M",  # minute, unpadded
    "ss": "%S",  # second, padded
    "s": "%-S",  # second, unpadded
    "yyyy": "%Y",  # year, full
    "yy": "%y",  # year, two-digit
    "y": "%Y",  # year, full
    "Z": "%Z",  # timezone
    "zzzz": "%Z",  # timezone
    "z": "%Z",  # timezone
    "SSSSSS": "%f",  # microseconds
}


def _match_ldml_to_posix(ldml: str) -> tuple[int, str] | None:
    for k in _ldml_to_posix:
        pos = ldml.find(k)
        if pos != -1:
            return pos, k

    return None


def _ldml_to_posix_format(ldml: str) -> str:
    res = _match_ldml_to_posix(ldml)
    if res is None:
        return ldml

    pos, k = res

    return (
        _ldml_to_posix_format(ldml[:pos])
        + _ldml_to_posix[k]
        + _ldml_to_posix_format(ldml[pos + len(k) :])
    )


@cache.function(id_arg_subset=["lang", "text"])
def _cached_translate(lang: str, text: str, translator: GoogleTranslator | None) -> str:
    translator = (
        translator
        if translator is not None
        and translator.source == "en"
        and translator.target == lang
        else GoogleTranslator(source="en", target=lang)
    )
    return translator.translate(text)


@dataclass
class Localization:
    """Locale config, which can be used as a context manager to activate the locale."""

    loc: Locale | None = None
    """Locale to use for this localization. If None, use parent context's locale."""

    overrides: Overrides | dict[Locale, Overrides] | None = None
    """Optional translation overrides for the current or other locales."""

    show_raw: bool = False
    """Whether to show raw function calls instead of localized text, for debugging."""

    def _machine_translate(self, text: str, locale: Locale | None = None) -> str:
        return _cached_translate(
            locale.language if locale is not None else self.locale.language,
            text,
            self.__translator,
        )

    @property
    def locale(self) -> Locale:
        """Return first locale found up the parent tree or default locale."""
        if self.loc is not None:
            return self.loc
        elif self.__parent is not None and self.__parent is not self:
            return self.__parent.locale

        return _get_default_locale()

    @staticmethod
    def _get_file_overrides(loc: Locale) -> Overrides:
        if file_overrides is None:
            return Overrides()

        return (
            file_overrides.get(str(loc))
            or file_overrides.get(str(loc.language))
            or Overrides()
        )

    def get_overrides(self, locale: Locale) -> Overrides:
        """Return merger of all the parents' and self's overrides for given locale."""
        parent_overrd = (
            self.__parent.get_overrides(locale)
            if self.__parent is not None and self.__parent is not self
            else self._get_file_overrides(locale)
        )
        self_overrd = (
            (self.overrides.get(locale) or Overrides())
            if isinstance(self.overrides, dict)
            else self.overrides
            if self.overrides is not None and self.locale == locale
            else Overrides()
        )

        return parent_overrd.merge(self_overrd)

    def __post_init__(self):  # noqa: D105
        self.__parent = None
        self.__token = None
        self.__translator = GoogleTranslator(source="en", target=self.locale.language)

    def activate(self):
        """Set this the as current localization."""
        self.__parent = active_localization.get()
        self.__token = active_localization.set(self)
        setlocale(LC_ALL, str(self.locale))

    def __enter__(self):  # noqa: D105
        self.activate()
        return self

    def deactivate(self, *_):
        """Reset locale to previous value."""
        if self.__token is not None:
            active_localization.reset(self.__token)
        if self.__parent is not None:
            setlocale(LC_ALL, str(self.__parent.locale))

    def __exit__(self, *_):  # noqa: D105
        self.deactivate()

    # Define custom deepcopy to avoid error when copying `self.__token`.
    def __deepcopy__(self, memo: dict) -> Self:  # noqa: D105
        c = Localization(
            loc=self.loc,
            overrides=self.overrides,
            show_raw=self.show_raw,
        )
        memo[id(self)] = c
        return cast(Self, c)

    def term(
        self,
        term: str,
        locale: Locale | None = None,
    ) -> str:
        """Localize given term.

        Args:
            term: Term to localize.
            locale:
                Locale to use for localization.
                Defaults to the currently activated locale.

        Returns:
            Localized term.
        """
        if self.show_raw:
            return f"term('{term}')"

        base = self.get_overrides(Locale("en", "US"))
        translation = self.get_overrides(locale or self.locale)

        if term in translation.vocabulary:
            return translation.vocabulary[term]
        elif term in base.vocabulary:
            return self._machine_translate(base.vocabulary[term], locale)

        return self._machine_translate(term, locale)

    @staticmethod
    def _get_template(overrides: Overrides, name: str, args: tuple) -> Template:
        texts = overrides.templates.get(name, {})
        texts = texts if isinstance(texts, dict) else {Args(all=True): texts}

        matches: dict[float, str | Template] = {}
        for match, tmpl in texts.items():
            matcher = (
                match
                if isinstance(match, Args)
                else Args(set(match) if isinstance(match, tuple) else match)
            )
            score = matcher.match(args)
            if score > 0:
                matches[score] = tmpl

        template = matches[max(matches.keys())] if len(matches) > 0 else None

        return (
            template
            if isinstance(template, Template)
            else Template(template)
            if template is not None
            else Template()
        )

    def text(
        self,
        context: str,
        content: str | None,
        *args: Any,
        locale: Locale | None = None,
    ) -> str:
        """Localize given text.

        Args:
            context: Label indentifying the context of the text to localize.
            content: Text to localize.
            args: Extra args to pass, if given text content is a template string.
            locale:
                Locale to use for localization.
                Defaults to the currently activated locale.

        Returns:
            Localized text.
        """
        if self.show_raw:
            arg_str = (", " + ", ".join(args)) if len(args) > 0 else ""
            return f"text('{content}'" + arg_str + f", context={context})"

        base = self.get_overrides(Locale("en", "US"))
        translation = self.get_overrides(locale or self.locale)

        tmpl = (
            self._machine_translate(content, locale) if content is not None else "{0}"
        )
        tmpl_args = [self.value(a, locale=locale) for a in args]

        transl_template = self._get_template(translation, context, args)
        base_template = self._get_template(base, context, args)

        if transl_template.replace is not None:
            tmpl = transl_template.replace
        elif base_template.replace is not None:
            tmpl = self._machine_translate(base_template.replace, locale)

        if transl_template.wrap is not None:
            tmpl = transl_template.wrap.format(tmpl)
        elif base_template.wrap is not None:
            tmpl = self._machine_translate(base_template.wrap, locale).format(tmpl)

        if transl_template.args is not None:
            if isinstance(transl_template.args, tuple):
                tmpl_args = transl_template.args
            elif isinstance(transl_template.args, dict):
                tmpl_args = {
                    i: transl_template.args[i] if i in transl_template.args else a
                    for i, a in enumerate(tmpl_args)
                }
        elif base_template.args is not None:
            if isinstance(base_template.args, tuple):
                tmpl_args = [self.value(a, locale=locale) for a in base_template.args]
            elif isinstance(base_template.args, dict):
                tmpl_args = {
                    i: self.value(base_template.args[i], locale=locale)
                    if i in base_template.args
                    else a
                    for i, a in enumerate(tmpl_args)
                }

        return tmpl.format(*tmpl_args)

    def label(
        self,
        label: str,
        context: str | None = None,
        locale: Locale | None = None,
    ) -> str:
        """Localize given label.

        Args:
            label: Label to localize.
            context: Context in which the label is used.
            locale:
                Locale to use for localization.
                Defaults to the currently activated locale.

        Returns:
            Localized label.
        """
        if self.show_raw:
            return f"label('{label}', context={context})"

        return (
            self.text(context, None, label, locale=locale)
            if context is not None
            else self.term(label, locale=locale)
        )

    def value(
        self,
        v: Any,
        options: Format = Format(),
        locale: Locale | None = None,
    ) -> str:
        """Return localized string represention of given value.

        Args:
            v: Value to localize.
            options: Options for formatting.
            context: Context in which the value is used.
            locale: Locale to use for localization.

        Returns:
            Localized value.
        """
        if self.show_raw:
            return f"value({v})"

        locale = locale or self.locale

        base = self.get_overrides(Locale("en", "US"))
        translation = self.get_overrides(locale or self.locale)

        options = base.merge(translation).format.merge(options)

        match (v):
            case _ if v is None or pd.isna(v):
                return options.na_representation
            case float() | Decimal() | int():
                fixed_digits = options.decimal_digits or options.decimal_min_digits or 0
                max_digits = options.decimal_digits or options.decimal_max_digits or 6
                group_sep = (
                    options.decimal_group_separator
                    if options.decimal_group_separator is not None
                    else False
                )

                match (options.decimal_notation):
                    case "percent":
                        return format_decimal(
                            v,
                            format=f"#,##0.{''.join(['0']*fixed_digits)}"
                            f"{''.join(['#']*(max_digits - fixed_digits))}%",
                            locale=locale,
                            group_separator=group_sep,
                        )
                    case "scientific":
                        return format_decimal(
                            v,
                            format=f"#,##0.{''.join(['0']*fixed_digits)}"
                            f"{''.join(['#']*(max_digits - fixed_digits))}E00",
                            locale=locale,
                            group_separator=group_sep,
                        )
                    case "plain" | _:
                        return format_decimal(
                            v,
                            format=f"#,##0.{''.join(['0']*fixed_digits)}"
                            f"{''.join(['#']*(max_digits - fixed_digits))}",
                            locale=locale,
                            group_separator=group_sep,
                        )
            case datetime() | pd.Timestamp():
                return format_datetime(
                    v,
                    options.datetime_format or options.datetime_auto_format or "medium",
                    locale=locale,
                )
            case date():
                return format_date(
                    v,
                    options.datetime_format or options.datetime_auto_format or "short",
                    locale=locale,
                )
            case time():
                return format_time(
                    v,
                    options.datetime_format or options.datetime_auto_format or "medium",
                    locale=locale,
                )
            case timedelta() | pd.Timedelta() if options.timedelta_auto_format != "raw":
                return format_timedelta(
                    v,
                    format=(
                        options.timedelta_auto_format
                        or options.datetime_auto_format
                        or "short"
                    ),
                    granularity=options.timedelta_resolution or "second",
                    threshold=0.99,
                    locale=locale,
                )
            case pd.Interval() if isinstance(v.left, pd.Timestamp) and isinstance(
                v.right, pd.Timestamp
            ):
                # Only do locale-aware formatting for datetime intervals.
                return format_interval(
                    v.left, v.right, skeleton=options.datetime_format, locale=locale
                )
            case Country():
                if (
                    options.country_format == GeoScheme.country_name
                    or options.country_format is None
                ):
                    return locale.territories.get(
                        str(v.to(GeoScheme.cc_iso2))
                    ) or self.term(
                        str(v.to(GeoScheme.country_name)),
                        locale=locale,
                    )
                else:
                    return str(v.to(options.country_format))
            case _:
                return (
                    self.term(str(v), locale=locale)
                    if options.fallback_to_translation
                    else str(v)
                )

    def formatter(
        self, options: Format = Format(), locale: Locale | None = None
    ) -> Callable[[Any], str]:
        """Return a formatting function bound to this locale.

        Args:
            options: Options for formatting.
            locale: Locale to use for localization.

        Returns:
            Formatting function for generating localized values.
        """
        return partial(self.value, options=options, locale=locale)

    def format_spec(
        self, typ: type, options: Format = Format(), locale: Locale | None = None
    ) -> str:
        """Return locale-aware format-string for given type, if applicable.

        Args:
            typ: Type to get format-string for.
            options: Options for formatting.
            locale: Locale to use for localization.
        """
        locale = locale or self.locale

        base = self.get_overrides(Locale("en", "US"))
        translation = self.get_overrides(locale or self.locale)

        options = base.merge(translation).format.merge(options)

        if issubclass(typ, int):
            match (options.decimal_notation):
                case "percent":
                    return ".0%"
                case "scientific":
                    return f".{options.decimal_min_digits or 6}e"
                case "plain" | _:
                    return "n"
        elif issubclass(typ, float | Decimal):
            match (options.decimal_notation):
                case "percent":
                    return f".{options.decimal_min_digits or 2}%"
                case "scientific":
                    return f".{options.decimal_min_digits or 6}e"
                case "plain" | _:
                    return f".{options.decimal_min_digits or 3}f"
        elif issubclass(typ, datetime | pd.Timestamp):
            opt = options.datetime_auto_format or "medium"
            return _ldml_to_posix_format(
                options.datetime_format
                or str.format(
                    locale.datetime_formats[opt],
                    locale.date_formats[opt],
                    locale.time_formats[opt],
                )
            )
        elif issubclass(typ, date):
            return _ldml_to_posix_format(
                options.datetime_format
                or locale.date_formats[options.datetime_auto_format or "medium"]
            )
        elif issubclass(typ, time):
            return _ldml_to_posix_format(
                options.datetime_format
                or locale.time_formats[options.datetime_auto_format or "medium"]
            )
        else:
            return ""


active_localization: ContextVar[Localization] = ContextVar(
    "active_localization", default=Localization()
)


def get_localization() -> Localization:
    """Get currently active localization."""
    return active_localization.get()


def iter_locales(
    locales: list[str],
    overrides: dict[str, Overrides] | None = None,
) -> Generator[Localization, None, None]:
    """Iterate over localizations for given locales w/ optional overrides.

    Args:
        locales: Locales to iterate over.
        overrides: Optional text overrides for the localizations.

    Returns:
        Generator of localizations.
    """
    for loc in locales:
        locz = Localization(
            loc=Locale.parse(loc, sep=("_" if "_" in loc else "-")),
            overrides={
                Locale.parse(k, sep=("_" if "_" in k else "-")): v
                for k, v in overrides.items()
            }
            if overrides is not None
            else None,
        )
        with locz:
            yield locz



"""Test enums module."""

from py_research.enums import StrEnum


class TestEnum(StrEnum):
    """Test enum."""

    __test__ = False

    VALUE1 = "value1"
    VALUE2 = "value2"


def test_str_enum_repr():
    """Test the __repr__ method of StrEnum."""
    assert repr(TestEnum.VALUE1) == "'value1'"
    assert repr(TestEnum.VALUE2) == "'value2'"



"""On-the-fly, file-based caching of function return values."""

import datetime
import json
import pickle
from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from typing import Any, ParamSpec, TypeVar, cast

import numpy as np
import pandas as pd
import yaml
from bs4 import BeautifulSoup, Tag

from py_research.data import gen_id
from py_research.files import ensure_dir_exists
from py_research.reflect import (
    get_calling_module_name,
    get_full_args_dict,
    get_return_type,
)
from py_research.telemetry import get_logger

log = get_logger()

default_root_path = Path.cwd() / ".cache"


P = ParamSpec("P")
R = TypeVar("R")


@dataclass
class FileCache:
    """Local, directory-based cache for storing function results."""

    path: Path
    """Root directory for storing cached results."""

    max_cache_time: datetime.timedelta = datetime.timedelta(days=7)
    """After how long to invalidate cached objects and recompute."""

    def __post_init__(self):  # noqa: D105
        now = datetime.datetime.now()
        self.__earliest_date = now - self.max_cache_time
        self.__now_str = now.strftime("%Y-%m-%d")

    @staticmethod
    def __get_date_from_filename(f: Path):
        return datetime.datetime.strptime(f.name.split(".")[0], "%Y-%m-%d")

    def __filter_outdated(self, f: Path):
        if self.__get_date_from_filename(f) > self.__earliest_date:
            return True
        else:
            f.unlink()
            return False

    def function(  # noqa: C901
        self,
        id_arg_subset: list[int] | list[str] | None = None,
        use_raw_arg: bool = False,
        id_callback: Callable[P, dict[str, Any] | None] | None = None,
        use_json: bool = True,
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator to cache wrapped function.

        Args:
            id_arg_subset:
                Number or name of the arguments to base hash id of result on.
            use_raw_arg:
                If ``True``, use the unhashed, string-formatted value of the id arg
                as filename. Only works for single id arg.
            id_callback:
                Callback function to use for retrieving a unique id from the arguments.
            use_json:
                Whether to use JSON as the format for caching dicts (instead of YAML).
        """

        def inner(func: Callable[P, R]) -> Callable[P, R]:  # noqa: C901
            path = ensure_dir_exists(self.path / func.__name__)

            @wraps(func)
            def inner_inner(*args: P.args, **kwargs: P.kwargs) -> R:  # noqa: C901
                id_args = None

                if id_callback is not None:
                    id_args = id_callback(*args, **kwargs)

                if id_args is None:
                    id_args = get_full_args_dict(func, args, kwargs)
                    if id_arg_subset is not None:
                        id_args = {
                            k: v for k, v in id_args.items() if k in id_arg_subset
                        }

                id_value = gen_id(
                    id_args if len(id_args) > 1 else list(id_args.values())[0],
                    raw_str=use_raw_arg,
                )

                filename_pattern = f"[0-9]*-[0-9]*-[0-9]*.{id_value}.*"

                result: R | None = None

                all_cached = [
                    f
                    for f in path.iterdir()
                    if f.match(filename_pattern) and self.__filter_outdated(f)
                ]

                if len(all_cached):
                    log.debug(f"💾 Taking cached result for: '{id_value}'")

                    last_cached = sorted(all_cached, key=self.__get_date_from_filename)[
                        -1
                    ]
                    extension = last_cached.name.split(".")[-1]
                    return_type = get_return_type(func) or type

                    cached_result = None
                    if issubclass(return_type, str) and extension == "txt":
                        with open(
                            last_cached,
                            encoding="utf-8",
                        ) as f:
                            cached_result = f.read()
                    elif issubclass(return_type, dict | list) and extension in (
                        "yaml",
                        "yml",
                    ):
                        cached_result = yaml.load(
                            open(last_cached, encoding="utf-8"), Loader=yaml.CLoader
                        )
                    elif issubclass(return_type, dict | list) and extension == "json":
                        cached_result = json.load(open(last_cached, encoding="utf-8"))
                    elif (
                        issubclass(return_type, pd.DataFrame | pd.Series)
                        and extension == "xlsx"
                    ):
                        cached_result = pd.read_excel(
                            last_cached, header=0, index_col=0
                        )
                    elif (
                        issubclass(return_type, pd.DataFrame | pd.Series)
                        and extension == "csv"
                    ):
                        cached_result = pd.read_csv(last_cached, header=0, index_col=0)
                    elif issubclass(return_type, np.ndarray) and extension == "npy":
                        cached_result = np.load(last_cached)
                    elif (
                        issubclass(return_type, BeautifulSoup | Tag)
                        and extension == "html"
                    ):
                        cached_result = BeautifulSoup(open(last_cached))
                    elif extension == "pkl":
                        cached_result = pickle.load(open(last_cached, mode="rb"))

                    result = cast(R | None, cached_result)

                if result is None:
                    log.debug(
                        f"⬇ Performing operation / fetching resource: '{id_value}'"
                    )

                    result = func(*args, **kwargs)

                    match (result):
                        case str():
                            with open(
                                path / f"{self.__now_str}.{id_value}.txt",
                                "w",
                                encoding="utf-8",
                            ) as f:
                                f.write(result)
                        case dict() | list():
                            if use_json:
                                json.dump(
                                    result,
                                    open(
                                        path / f"{self.__now_str}.{id_value}.json",
                                        "w",
                                        encoding="utf-8",
                                    ),
                                    indent=2,
                                )
                            else:
                                yaml.dump(
                                    result,
                                    open(
                                        path / f"{self.__now_str}.{id_value}.yaml",
                                        "w",
                                        encoding="utf-8",
                                    ),
                                    allow_unicode=True,
                                )
                        case pd.DataFrame():
                            result.to_excel(
                                path / f"{self.__now_str}.{id_value}.xlsx", index=True
                            )
                        case np.ndarray():
                            np.save(path / f"{self.__now_str}.{id_value}.npy", result)
                        case BeautifulSoup() | Tag():
                            with open(
                                path / f"{self.__now_str}.{id_value}.html", mode="w"
                            ) as file:
                                file.write(str(result))
                        case _:
                            pickle.dump(
                                result,
                                open(
                                    path / f"{self.__now_str}.{id_value}.pkl", mode="wb"
                                ),
                            )

                return result

            return inner_inner

        return inner


def get_cache(
    name: str | None = None,
    root_path: Path | None = None,
    max_cache_time: datetime.timedelta = datetime.timedelta(days=7),
):
    """Return a named cache instance private to the calling module.

    Args:
        name: Name of the cache (directory) to create.
        root_path: Root directory, where to store cache.
        max_cache_time: After how many days to invalidate cached objects and recompute.

    Returns:
        A cache instance.
    """
    root_path = root_path or default_root_path
    calling_module = get_calling_module_name() or "root"

    return FileCache(
        ensure_dir_exists(root_path / calling_module / (name or "")),
        max_cache_time,
    )



"""Easy to use relational database."""

from .base import DB as DB
from .base import DBSchema as DBSchema
from .base import Table as Table
from .conflicts import DataConflictError as DataConflictError
from .importing import tree_to_db as tree_to_db



# pylint: disable=C0415:import-outside-toplevel
# pylint: disable=E1136:unsubscriptable-object
# pylint: disable=W0622:redefined-builtin
"""Utilities for monitoring long-running functions via logging, tracing and metrics."""

import logging
import sys
from collections.abc import Iterable, Mapping
from functools import wraps
from logging import StreamHandler
from typing import IO, Literal, TypeVar, cast

import structlog
from stqdm import stqdm
from tqdm import tqdm as base_tqdm
from tqdm.autonotebook import tqdm as atqdm


class TqdmHandler(StreamHandler):
    """A handler class which allows the cursor to stay on one line."""

    tqdm: base_tqdm | None = None

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(*args, **kwargs)

    def emit(self, record):  # noqa: D102
        try:
            msg = self.format(record)

            if self.tqdm is not None:
                self.tqdm.set_postfix_str(msg)
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)


def _check_streamlit():
    """Check whether python code is run within streamlit."""
    try:
        from streamlit.runtime.scriptrunner import get_script_run_ctx

        if not get_script_run_ctx():
            use_streamlit = False
        else:
            use_streamlit = True
    except ModuleNotFoundError:
        use_streamlit = False
    return use_streamlit


T = TypeVar("T")


@wraps(base_tqdm)
def tqdm(
    iterable: Iterable[T],
    desc: str | None = None,
    total: float | None = None,
    leave: bool | None = True,
    file: IO[str] | None = None,
    ncols: int | None = None,
    mininterval: float = 0.1,
    maxinterval: float = 10.0,
    miniters: float | None = None,
    ascii: bool | str | None = None,
    disable: bool | None = False,
    unit: str = "it",
    unit_scale: bool | float = False,
    dynamic_ncols: bool = False,
    smoothing: float = 0.3,
    bar_format: str | None = None,
    initial: float = 0,
    position: int | None = None,
    postfix: Mapping[str, object] | str | None = None,
    unit_divisor: float = 1000,
    write_bytes: bool = False,
    lock_args: tuple[bool | None, float | None] | tuple[bool | None] | None = None,
    nrows: int | None = None,
    colour: str | None = None,
    delay: float | None = 0,
    gui: bool = False,
    **other_kwargs,
) -> Iterable[T]:
    """Return a tqdm instace adapted to the current environment.

    (Terminal, Jupyter or Streamlit)
    """
    kwargs = dict(
        iterable=iterable,
        desc=desc,
        total=total,
        leave=leave,
        file=file,
        ncols=ncols,
        mininterval=mininterval,
        maxinterval=maxinterval,
        miniters=miniters,
        ascii=ascii,
        disable=disable,
        unit=unit,
        unit_scale=unit_scale,
        dynamic_ncols=dynamic_ncols,
        smoothing=smoothing,
        bar_format=bar_format,
        initial=initial,
        position=position,
        postfix=postfix,
        unit_divisor=unit_divisor,
        write_bytes=write_bytes,
        lock_args=lock_args,
        nrows=nrows,
        colour=colour,
        delay=delay,
        gui=gui,
        **other_kwargs,
    )

    res_tqdm = stqdm(**kwargs) if _check_streamlit() else atqdm(**kwargs)  # type: ignore  # noqa: E501
    TqdmHandler.tqdm = res_tqdm
    return cast(Iterable[T], res_tqdm)


def configure_logging(purpose: Literal["status", "report", "log"] = "status") -> None:
    """Auto-configure :py:mod:`structlog` based on logging purpose.

    Args:
        purpose:
            Which purpose the default global logger is supposed to fulfill.
            Can be any of:

            * ``status``:
                Inform about the current status of long-running functions
                via a single, changing console line.
            * ``report``:
                Record important steps / intermediate values / logical branches
                of large functions to the console.
            * ``log``:
                Produce a common log stream in JSON format.
    """
    structlog.configure(
        processors=[
            # If log level is too low, abort pipeline and throw away log entry.
            structlog.stdlib.filter_by_level,
            # Perform %-style formatting.
            structlog.stdlib.PositionalArgumentsFormatter(),
            *(
                [  # Add the name of the logger to event dict.
                    structlog.stdlib.add_logger_name,
                    # Add log level to event dict.
                    structlog.stdlib.add_log_level,
                    # Add a timestamp in ISO 8601 format.
                    structlog.processors.TimeStamper(fmt="iso"),
                    # If the "stack_info" key in the event dict is true, remove it and
                    # render the current stack trace in the "stack" key.
                    structlog.processors.StackInfoRenderer(),
                    # If the "exc_info" key in the event dict is either true or a
                    # sys.exc_info() tuple, remove "exc_info" and render the exception
                    # with traceback into the "exception" key.
                    structlog.processors.format_exc_info,
                    # If some value is in bytes, decode it to a unicode str.
                    structlog.processors.UnicodeDecoder(),
                    # Add callsite parameters.
                    structlog.processors.CallsiteParameterAdder(
                        {
                            structlog.processors.CallsiteParameter.FILENAME,
                            structlog.processors.CallsiteParameter.FUNC_NAME,
                            structlog.processors.CallsiteParameter.LINENO,
                        }
                    ),
                    structlog.dev.ConsoleRenderer()
                    if purpose != "log"
                    else structlog.processors.JSONRenderer(),
                ]
                if purpose != "status"
                else [structlog.dev.ConsoleRenderer(colors=False)]
            ),
        ],
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.stdlib.BoundLogger,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
        # string) from the final processor (`JSONRenderer`) will be passed to
        # the method of the same name as that you've called on the bound logger.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Effectively freeze configuration after creating the first bound
        # logger.
        cache_logger_on_first_use=True,
    )

    logging.basicConfig(
        format="%(message)s",
        handlers=[
            TqdmHandler(sys.stdout)
            if purpose == "status"
            else StreamHandler(sys.stdout)
        ],
        level=logging.DEBUG if purpose == "status" else logging.INFO,
        force=True,
    )


@wraps(structlog.get_logger)
def get_logger(name: str | None = None, **kwds) -> structlog.stdlib.BoundLogger:
    """Typed interface for `structlog.get_logger`."""
    return structlog.get_logger(name, **kwds)


# Ensure that logging is configured to use `structlog.stdlib.BoundLogger`
# for `get_logger()`.
configure_logging()



"""Utilities for creating pretty result tables."""

from collections.abc import Callable
from dataclasses import InitVar, dataclass, field
from functools import reduce
from pathlib import Path
from typing import Any, Literal, cast

import imgkit
import pandas as pd
import pdfkit
from pandas.api.types import is_float_dtype, is_integer_dtype, is_numeric_dtype
from pandas.io.formats.style import Styler


@dataclass
class TableStyle:
    """Define a pretty table column format."""

    cols: str | tuple[str, str] | list[str | tuple[str, str]] | None = None
    """Column(s) to apply this style to. If None, apply to all columns.
    Index levels can be styled via their names as if they were columns.
    """

    rows: pd.Series | None = None
    """Rows to apply this style to. If None, apply to all rows."""

    name: str | None = None
    """Name of this style"""

    str_format: str | Callable[[Any], str] | None = None
    """Format string or callback function to use for formatting numerical values"""

    alignment: Literal["left", "center", "right"] | None = None
    """How to align the columns' text"""

    css: dict[str, str] | Callable[[Any], str] | None = None
    """Custom CSS styles to apply to matched cells"""

    hide_headers: bool | None = None
    """Whether to hide headers of given ``cols``."""

    hide_rows: bool | None = None
    """Whether to hide cols/rows if matched by this selection."""

    filter_inclusive: bool | None = None
    """Whether to show cols/rows if matched by this selection."""

    filter_exclusive: bool | None = None
    """Whether to only show cols/rows if matched by this selection."""


@dataclass
class TableColors:
    """Define colors for a pretty table."""

    row_even: str = "#f2f2f2"
    """Background color for even rows."""

    row_odd: str = "#fff"
    """Background color for odd rows."""

    header_even: str = "#677a96"
    """Background color for even headers."""

    header_odd: str = "#677a96"
    """Background color for odd headers."""


@dataclass
class ResultTable:
    """Define and render a pretty result table with custom formatting and highlights."""

    df: InitVar[pd.DataFrame]
    """Dataframe to render as pretty table."""

    styles: list[TableStyle] = field(default_factory=list)
    """Styles to apply to the table."""

    labels: dict[str, str] | Callable[[str], str] = field(default_factory=dict)
    """Labels to use for the table headers.

    If dict, the keys are column names and the values are the labels.
    If callable, the function is called with the column name and should return the label
    """

    widths: dict[str, float | str] = field(default_factory=dict)
    """Widths to use for the table columns.

    If a float, the width is relative to the sum of all widths.
    If a string, the width is set to the CSS string value.
    """

    title: str | None = None
    """Title of the table."""

    hide_index: bool | list[str] = True
    """Whether to hide the index columns."""

    max_row_cutoff: int = 100
    """Maximum number of rows to render."""

    font_size: float = 1.0
    """Default font size to use for the table."""

    default_style: TableStyle = field(default_factory=TableStyle)
    """Default style to apply to the table."""

    column_flatten_format: str | None = None
    """Format string to use for flatteting multi-index column labels.
    Leave as None to keep multi-index column labels as tuples.
    Format string must take two positional arguments, e.g. "{0}_{1}".
    """

    table_colors: TableColors = field(default_factory=TableColors)
    """Colors to use for this table."""

    table_styles: dict[str, dict[str, str]] = field(default_factory=dict)
    """Additional styles to apply to the table.
    Dictionary keys must be CSS selectors and
    values must be dictionaries of CSS properties.
    """

    def __post_init__(self, df: pd.DataFrame):  # noqa: D105
        if self.hide_index is not True:
            hidden_indexes = (
                self.hide_index if isinstance(self.hide_index, list) else []
            )
            index_names = (
                df.index.names
                if all(df.index.names)
                else [f"index_{i}" for i in range(df.index.nlevels)]
            )

            df = df.rename_axis(index=index_names)

            index_col_names = [
                name for name in index_names if name not in hidden_indexes
            ]
            if df.columns.nlevels > 1:
                index_col_names = [("", name) for name in index_col_names]

            df = df.copy()
            for col_name in index_col_names:
                index_name = col_name if isinstance(col_name, str) else col_name[1]
                df[col_name] = df.index.get_level_values(index_name)

            self.data = df[
                [*index_col_names, *[c for c in df.columns if c not in index_col_names]]
            ]
        else:
            self.data = df

    @property
    def __hidden_headers(self) -> set[str | tuple[str, str]]:
        hide_styles = [
            set([style.cols] if isinstance(style.cols, str) else style.cols)
            for style in self.styles
            if style.cols is not None and style.hide_headers is True
        ]
        return (
            reduce(
                lambda x, y: x | y,
                hide_styles,
            )
            if len(hide_styles) > 0
            else set()
        )

    def html_description(self, full_doc: bool = True) -> str:
        """Return HTML description for table filters and highlights.

        Args:
            full_doc: Whether to wrap the description in a full HTML document.

        Returns:
            HTML code for the description.
        """
        highlights = [
            (
                h,
                "; ".join(f"{prop}: {val}" for prop, val in h.css.items())
                if isinstance(h.css, dict)
                else str(getattr(h.css, "__name__"))
                if hasattr(h.css, "__name__")
                else None,
            )
            for h in self.styles
        ]

        exclusive_h = [
            (h, css)
            for h, css in highlights
            if h.name is not None and css is not None and h.filter_exclusive
        ]
        inclusive_h = [
            (h, css)
            for h, css in highlights
            if h.name is not None and css is not None and h.filter_inclusive
        ]
        highlight_h = [
            (h, css)
            for h, css in highlights
            if h.name is not None
            and css is not None
            and not any([h.filter_exclusive, h.filter_inclusive, h.hide_rows])
        ]
        hide_h = [h for h, _ in highlights if h.name is not None and h.hide_rows]

        desc = "\n".join(
            [
                f"""<h2>Show rows where all of:</h2>
            <ul>
                {
                    ''.join(
                        [
                            '<li style="margin-bottom: 0.5rem;">'
                            + f'<span style="display: inline-block; {css};">'
                            + css
                            + '</span>:'
                            + f' {h.name}'
                            + '</li>'
                            for h, css in exclusive_h
                        ]
                    )
                }
            </ul>
            """
                if len(exclusive_h) > 0
                else "",
                f"""<h2>Show rows where any of:</h2>
            <ul>
                {
                    ''.join(
                        [
                            '<li style="margin-bottom: 0.5rem;">'
                            + f'<span style="display: inline-block; {css};">'
                            + css
                            + '</span>:'
                            + f' {h.name}'
                            + '</li>'
                            for h, css in inclusive_h
                        ]
                    )
                }
            </ul>
            """
                if len(inclusive_h) > 0
                else "",
                f"""<h2>Highlight rows where:</h2>
            <ul>
                {
                    ''.join(
                        [
                            '<li style="margin-bottom: 0.5rem;">'
                            + f'<span style="display: inline-block; {css};">'
                            + css
                            + '</span>:'
                            + f' {h.name}'
                            + '</li>'
                            for h, css in highlight_h
                        ]
                    )
                }
            </ul>
            """
                if len(highlight_h) > 0
                else "",
                f"""<h2>Hide rows where any of:</h2>
            <ul>
                {
                    ''.join(
                        [
                            '<li style="margin-bottom: 0.5rem;">'
                            + f'{h.name}'
                            + '</li>'
                            for h in hide_h
                        ]
                    )
                }
            </ul>
            """
                if len(hide_h) > 0
                else "",
            ]
        )

        return (
            f"""
        <!doctype html>
        <html>
            <head>
                <title>{self.title} - highlight-description</title>
            </head>
            <body style="font-family: sans-serif">
                <h1>Description for table '{self.title}'</h1>
                {desc}
            </body>
        </html>
        """
            if full_doc
            else desc
        )

    def _get_cols_by_second_level(
        self, cols: list[str | tuple[str, str]]
    ) -> list[tuple[str, str]]:
        return [
            cast(tuple[str, str], cc)
            for c in cols
            for cc in self.data.columns
            if (isinstance(c, str) and cc[1] == c) or cc == c
        ]

    def _to_styler_col(self, col: str | tuple[str, str]) -> str | tuple[str, str]:
        return (
            col
            if isinstance(col, str)
            or self.data.columns.nlevels == 1
            or self.column_flatten_format is None
            else self.column_flatten_format.format(*col)
        )

    def _default_str_format(self, col: str) -> str:
        all_col_data = (
            self.data.loc[
                :,
                self._get_cols_by_second_level([col]),
            ]
            if self.data.columns.nlevels > 1
            else self.data[[col]]
        )
        if all(is_integer_dtype(col_data) for _, col_data in all_col_data.items()):
            return "{:d}"
        elif all(is_float_dtype(col_data) for _, col_data in all_col_data.items()):
            if all(
                col_data.map(float.is_integer).all()
                for _, col_data in all_col_data.items()
            ):
                return "{:.0f}"
            else:
                return "{:.3f}"
        else:
            return "{}"

    def _default_alignment(self, col: str) -> str:
        all_col_data = (
            self.data.loc[
                :,
                self._get_cols_by_second_level([col]),
            ]
            if self.data.columns.nlevels > 1
            else self.data[[col]]
        )
        return (
            "right"
            if all(is_numeric_dtype(col_data) for _, col_data in all_col_data.items())
            else "left"
        )

    def _prettify_df(self, table: pd.DataFrame, font_size: float = 1.0) -> Styler:
        """Apply styles to a DataFrame or Styler and make it pretty."""
        return table.style.set_table_styles(
            [
                {
                    "selector": "",
                    "props": (
                        "font-family: Arial, Helvetica, sans-serif;"
                        "border-collapse: collapse; width: 100%;"
                        "color: black; text-align: left;"
                    ),
                },
                {
                    "selector": "caption",
                    "props": (
                        f"margin-bottom: 1rem;" f"font-size: {(font_size*1.5):.3g}rem;"
                    ),
                },
                {
                    "selector": "td, th",
                    "props": (
                        f"font-size: {font_size:.3g}rem;"
                        "border: 1px solid #ddd;"
                        "padding: 8px;"
                    ),
                },
                {
                    "selector": "tr:nth-child(even)",
                    "props": f"background-color: {self.table_colors.row_even};",
                },
                {
                    "selector": "tr:nth-child(odd)",
                    "props": f"background-color: {self.table_colors.row_odd};",
                },
                {
                    "selector": "th",
                    "props": (
                        "padding: 12px 8px 12px;" "text-align: left;" "color: black;"
                    ),
                },
                {
                    "selector": "th:nth-child(even)",
                    "props": f"background-color: {self.table_colors.header_even};",
                },
                {
                    "selector": "th:nth-child(odd)",
                    "props": f"background-color: {self.table_colors.header_odd};",
                },
                *(
                    {
                        "selector": selector,
                        "props": " ".join(
                            [
                                prop_name
                                + ": "
                                + prop_value.rstrip(" ").rstrip(";")
                                + " !important;"
                                for prop_name, prop_value in props.items()
                            ]
                        ),
                    }
                    for selector, props in self.table_styles.items()
                ),
            ]
        ).format(na_rep="")

    def _apply_default_style(self, styled: Styler) -> Styler:
        if isinstance(self.default_style.css, dict):
            styled = styled.set_properties(
                subset=None,
                **self.default_style.css,
            )
        elif isinstance(self.default_style.css, Callable):
            styled = styled.applymap(
                func=self.default_style.css,
            )

        return styled

    def _apply_col_defaults(self, styled: Styler) -> Styler:
        res = styled
        for col in self.data.columns:
            res = res.set_properties(
                subset=[self._to_styler_col(col)],
                **{
                    "text-align": self.default_style.alignment
                    or self._default_alignment(col)
                },
            )
            res = res.format(
                subset=[self._to_styler_col(col)],
                formatter=self.default_style.str_format
                or self._default_str_format(col),
            )

        return res

    def _apply_styles(self, styled: Styler) -> Styler:
        for style in self.styles:
            rows, cols = (
                style.rows if style.rows is not None else slice(None),
                style.cols
                if isinstance(style.cols, list)
                else [style.cols]
                if style.cols is not None
                else slice(None),
            )

            if isinstance(cols, list) and self.data.columns.nlevels > 1:
                cols = [
                    self._to_styler_col(c) for c in self._get_cols_by_second_level(cols)
                ]

            subset = (rows, cols)

            if style.hide_rows is not None:
                styled = styled.hide(
                    subset[0],  # type: ignore
                    axis="index",
                )
                continue

            if style.alignment is not None:
                styled = styled.set_properties(
                    subset=subset,  # type: ignore
                    **{"text-align": style.alignment},
                )

            if isinstance(style.css, dict):
                styled = styled.set_properties(
                    subset=subset,  # type: ignore
                    **style.css,
                )
            elif isinstance(style.css, Callable):
                styled = styled.applymap(
                    subset=subset,  # type: ignore
                    func=style.css,
                )

            if style.str_format is not None:
                styled = styled.format(
                    subset=subset,  # type: ignore
                    formatter=f"{{:{style.str_format}}}"
                    if isinstance(style.str_format, str)
                    else style.str_format,
                )

        return styled

    def __apply_widths(self, styled: Styler) -> Styler:
        widths = {
            c: (self.widths.get(c if self.data.columns.nlevels == 1 else c[1]) or 1)
            for c in self.data.columns
        }
        width_sum = sum(w for w in widths.values() if isinstance(w, int | float))
        for col, width in widths.items():
            styled = styled.set_properties(
                subset=(slice(None), [self._to_styler_col(col)]),  # type: ignore
                width=(
                    f"{(width / width_sum * 100):.1f}%"
                    if isinstance(width, int | float)
                    else width
                ),
            )

        return styled

    def __apply_labels(self, styled: Styler) -> Styler:
        label_func = self.labels.get if isinstance(self.labels, dict) else self.labels
        labels = (
            {
                c: "" if c in self.__hidden_headers else label_func(c) or c
                for c in self.data.columns
            }
            if self.data.columns.nlevels == 1
            else {
                c: ("", "")
                if c in self.__hidden_headers
                else (c[0], label_func(c[1]) or c[1])
                if c[1]
                else ("", label_func(c[0]) or c[0])
                for c in self.data.columns
            }
        )

        if self.column_flatten_format is not None and self.data.columns.nlevels > 1:
            labels = {
                c: self.column_flatten_format.format(*label)
                if c[1] not in self.data.index.names
                else label[1]
                for c, label in labels.items()
            }

        return styled.relabel_index(
            list(labels.values()),  # type: ignore
            axis="columns",
        )

    def to_styled_df(self) -> Styler:
        """Render table to styled pandas dataframe."""
        data = self.data.copy()

        if self.column_flatten_format is not None and self.data.columns.nlevels > 1:
            data.columns = [self.column_flatten_format.format(*c) for c in data.columns]

        incl_filters = [
            style.rows
            for style in self.styles
            if style.filter_inclusive is True and style.rows is not None
        ]
        if len(incl_filters) > 0:
            data = data.loc[reduce(lambda x, y: x | y, incl_filters)]

        excl_filters = [
            style.rows
            for style in self.styles
            if style.filter_exclusive is True and style.rows is not None
        ]
        if len(excl_filters) > 0:
            data = data.loc[reduce(lambda x, y: x & y, excl_filters)]

        styled = self._prettify_df(data.iloc[: self.max_row_cutoff], self.font_size)
        styled = self._apply_default_style(styled)
        styled = self._apply_col_defaults(styled)
        styled = self.__apply_widths(styled)
        styled = self.__apply_labels(styled)
        styled = self._apply_styles(styled)

        if self.title is not None:
            styled = styled.set_caption(self.title)

        styled = styled.hide(axis="index")

        return styled


def to_html(styled: Styler, full_doc: bool = True) -> str:
    """Return HTML representation of a pretty table.

    Args:
        styled: Styled dataframe to render.
        full_doc: Whether to wrap the table in a full HTML document.

    Returns:
        HTML code for the table.
    """
    return (
        f"""
        <!doctype html>
        <html>
            <head>
                <title>{getattr(styled, "caption") or ""}</title>
            </head>
            <body>
                {styled.to_html(escape=False)}
            </body>
        </html>
        """
        if full_doc
        else styled.to_html(escape=False)
    )


def html_to_pdf(doc: str, file: Path):
    """Render and save HTML ``doc`` as PDF document.

    Args:
        doc: HTML document to render.
        file: File to save the rendered PDF document to.
    """
    pdfkit.from_string(doc, file)


def html_to_image(doc: str, file: Path):
    """Render and save HTML ``doc`` as PNG image.

    Args:
        doc: HTML document to render.
        file: File to save the rendered image to.
    """
    imgkit.from_string(doc, file, options={"zoom": "3.125", "width": "3125"})



"""Utilities for working with geographical data, esp. data associated with countries."""

from collections.abc import Iterable
from dataclasses import dataclass
from enum import auto
from typing import Literal, TypeAlias, cast

import country_converter as coco
import pandas as pd
from typing_extensions import Self

from py_research.enums import StrEnum


class GeoAlliance(StrEnum):
    """List of international alliances used to define geo-regions of interest."""

    EU = auto()
    """`See Eurostat definition <https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:EU_enlargements>`_"""

    EU12 = auto()
    EU15 = auto()
    EU25 = auto()
    EU27 = auto()
    EU27_2007 = auto()
    EU28 = auto()

    EEA = auto()
    """`EEA according to Eurostat <https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:European_Economic_Area_(EEA)>`_"""

    G7 = auto()
    """`G7 according to Wikipedia <https://en.wikipedia.org/wiki/G7>`_"""

    G20 = auto()
    """`G20 according to Wikipedia <https://en.wikipedia.org/wiki/G20>`_"""

    APEC = auto()
    """`APEC according to Wikipedia <https://en.wikipedia.org/wiki/Asia-Pacific_Economic_Cooperation>`_"""

    BRIC = auto()
    """`BRIC according to Wikipedia <https://en.wikipedia.org/wiki/BRIC>`_"""

    BASIC = auto()
    """`BASIC according to Wikipedia <https://en.wikipedia.org/wiki/BASIC_countries>`_"""

    CIS = auto()
    """`CIS according to Wikipedia <https://en.wikipedia.org/wiki/Commonwealth_of_Independent_States>`_"""

    OECD = auto()
    """`OECD members <https://www.oecd.org/about/members-and-partners/>`_"""


class GeoScheme(StrEnum):
    """List of schemes, which can be used to define geo-regions of interest."""

    country_name = auto()
    """Short name of a country."""

    continent = auto()
    """Name of a continent."""

    cc_iso3 = auto()
    """ISO3 code of a country."""

    cc_iso2 = auto()
    """ISO2 code of a country."""

    alliance = auto()
    """Name of an international alliance, to which a country belongs.
    """

    world = auto()
    """Dummy scheme to match all of the world."""

    @staticmethod
    def _scheme_map() -> dict["GeoScheme", str]:
        return {
            GeoScheme.country_name: "name_short",
            GeoScheme.continent: "continent",
            GeoScheme.cc_iso3: "ISO3",
            GeoScheme.cc_iso2: "ISO2",
        }

    def to_coco_scheme(self) -> str | None:
        """Return associated coco scheme, if applicable."""
        return self._scheme_map().get(self)


@dataclass(frozen=True)
class GeoRegion:
    """Define a geo-region according to some scheme (country, continent, etc.)."""

    label: str
    """The geo-location's label according to `scheme`."""

    scheme: GeoScheme = GeoScheme.country_name
    """The naming/classification scheme used."""

    display_label: str | None = None
    """Optional, custom display label for the geo-location."""

    exclude_already_covered: bool = True
    """When listing multiple geo-regions, exclude locations
    which have already been covered by previously listed regions."""

    def get_label(self) -> str:
        """Return string label of this region."""
        return self.display_label or self.label

    def to_country_list(
        self,
        scheme: "CountryScheme" = GeoScheme.cc_iso3,
    ) -> list[str]:
        """Return list of matching countries in given ``scheme``.

        Args:
            scheme: The scheme to convert to countries.

        Returns:
            List of countries in ``scheme``.
        """
        res = None

        match (self.scheme):
            case (
                GeoScheme.country_name
                | GeoScheme.continent
                | GeoScheme.cc_iso3
                | GeoScheme.cc_iso2
            ):
                res = coco.convert(
                    self.label,
                    src=self.scheme.to_coco_scheme(),
                    to=scheme.to_coco_scheme(),
                )
            case GeoScheme.alliance:
                res = coco.convert(
                    self.label, src=self.label, to=scheme.to_coco_scheme()
                )
            case GeoScheme.world:
                res = coco.CountryConverter().data[scheme.to_coco_scheme()].to_list()
            case _:
                raise ValueError(f"Unsupported geo-location scheme: {self.scheme}")

        return res if isinstance(res, list) else [res]


def _region_to_country_map(
    region: GeoRegion, scheme: "CountryScheme" = GeoScheme.cc_iso3, cc_col: str = "cc"
) -> pd.DataFrame:
    """Resolve geo-region definition to a dataframe.

    Result df maps the region's (display) label to one or more
    ISO3 country codes.
    """
    iso3_list = region.to_country_list(scheme=scheme)
    return pd.DataFrame({cc_col: iso3_list}).assign(geo_region=region.get_label())


CountryScheme: TypeAlias = Literal[
    GeoScheme.country_name, GeoScheme.cc_iso2, GeoScheme.cc_iso3
]
"""A :py:class:`GeoScheme` which can be used to define a single country.`"""


@dataclass(frozen=True)
class Country:
    """A country represented by ISO2 code, ISO3 code or name."""

    label: str
    """The country's label according to `scheme`."""

    scheme: CountryScheme | None = None
    """The naming/classification scheme used."""

    def to(self, scheme: CountryScheme) -> Self:
        """Convert to other scheme."""
        return cast(
            Self,
            Country(
                cast(
                    str,
                    coco.convert(
                        self.label, to=scheme.to_coco_scheme(), src=self.scheme
                    ),
                ),
                scheme,
            )
            if scheme != self.scheme
            else self,
        )

    def __str__(self) -> str:  # noqa: D105
        return self.label

    def __format__(self, spec: str) -> str:  # noqa: D105
        return str(self.to(cast(CountryScheme, GeoScheme(spec))))


def countries_to_scheme(
    countries: pd.Series,
    target: CountryScheme = GeoScheme.cc_iso3,
    src: CountryScheme | None = None,
) -> pd.Series:
    """Translate given series of country labels to ``scheme``.

    Args:
        countries: Series of country labels.
        target: Target scheme to translate to.
        src: Source scheme to translate from.

    Returns:
        Series of translated country labels.
    """
    return pd.Series(
        coco.convert(
            countries,
            src=src.to_coco_scheme() if src is not None else None,
            to=target.to_coco_scheme(),
        ),
        index=countries.index,
    )


def expand_geo_col_to_cc(
    df: pd.DataFrame,
    geo_col: str,
    scheme: GeoScheme = GeoScheme.country_name,
    cc_scheme: CountryScheme = GeoScheme.cc_iso3,
) -> pd.DataFrame:
    """Expand geo-regions present in ``geo_col`` to country codes.

    Expand such that rows of ``df`` with multiple mapped CCs are multiplicated.

    Args:
        df: The dataframe to expand.
        geo_col: The column containing geo-regions.
        scheme: The scheme used to define the geo-regions.
        cc_scheme: The scheme to expand to.

    Returns:
        The expanded dataframe.
    """
    if len(df) == 0:
        return pd.DataFrame(columns=[geo_col])

    cc_col = "cc" if geo_col != "cc" else "cc_mapped"

    cc_map = pd.concat(
        [
            _region_to_country_map(
                GeoRegion(g, scheme=scheme), scheme=cc_scheme, cc_col=cc_col
            )
            for g in df[geo_col].dropna().unique()
        ]
    )

    return df.merge(cc_map, left_on=geo_col, right_on="geo_region", how="left").drop(
        columns="geo_region"
    )


def merge_geo_regions(
    df: pd.DataFrame,
    geo_col: str,
    geo_regions: Iterable[GeoRegion | str],
    input_scheme: GeoScheme = GeoScheme.country_name,
    pretty_labels: bool = True,
) -> pd.DataFrame:
    """Right-merge ``geo_regions`` onto ``df`` based on ``geo_col``.

    Merge such that rows with multiple mapped regions are multiplicated.

    Args:
        df: The dataframe to merge into.
        geo_col: The column containing geo-regions.
        geo_regions: The geo-regions to merge.
        input_scheme: The scheme used to define the geo-regions.
        rest_of_world: Whether to add a "Rest of World" region.
        pretty_labels: Whether to use pretty labels for regions.

    Returns:
        The merged dataframe.
    """
    src_df = expand_geo_col_to_cc(df, geo_col=geo_col, scheme=input_scheme)

    cc_coverage = set()
    res_df = pd.DataFrame()

    geo_regions = [
        *geo_regions,
        GeoRegion("world", GeoScheme.world, exclude_already_covered=True),
    ]

    for gr in geo_regions:
        gr = gr if isinstance(gr, GeoRegion) else GeoRegion(gr)
        geo_region_res = gr.get_label()

        cc_map = _region_to_country_map(gr)
        cc_set = set(cc_map["cc"])
        cc_already_covered = list(cc_coverage & cc_set)

        if pretty_labels and (
            gr.scheme == GeoScheme.cc_iso2 or gr.scheme == GeoScheme.cc_iso3
        ):
            geo_region_res = coco.convert(
                geo_region_res, src=gr.scheme.to_coco_scheme(), to="name"
            )

        if len(cc_already_covered) > 0 and gr.exclude_already_covered:
            geo_region_res = f"Rest of {gr.get_label()}"
            cc_map = cc_map.loc[~cc_map["cc"].isin(cc_already_covered)]

        res_df = pd.concat(
            [
                res_df,
                src_df.merge(cc_map, on="cc", how="inner").assign(
                    geo_region=geo_region_res, geo_region_label=gr.get_label()
                ),
            ]
        )
        cc_coverage |= cc_set

    return res_df


def match_to_geo_region(
    countries: pd.Series,
    geo_region: GeoRegion,
    country_scheme: CountryScheme | None = None,
) -> pd.Series:
    """Check whether countries are in given geo-region.

    Args:
        countries: Series of countries to check.
        geo_region: The geo-region to check against.
        country_scheme: The scheme of the countries.

    Returns:
        Series of booleans indicating whether countries are in geo-region.
    """
    return countries_to_scheme(countries, GeoScheme.cc_iso3, src=country_scheme).isin(
        geo_region.to_country_list(scheme=GeoScheme.cc_iso3)
    )


flag_sizes = pd.Series([20, 40, 80, 160, 320, 640, 1280, 2560])


def gen_flag_url(cc: pd.Series, width: int) -> pd.Series:
    """Get the URL of a small flag image for a given country code.

    Args:
        cc: Series of country codes.
        width: The desired width of the flag.

    Returns:
        Series of flag image URLs.
    """
    return (
        "https://flagcdn.com/w"
        + str(flag_sizes.loc[flag_sizes > width].min())
        + "/"
        + countries_to_scheme(cc, GeoScheme.cc_iso2).str.lower()
        + ".png"
    )


def gen_flag_img_tag(cc: pd.Series, width: int) -> pd.Series:
    """Generate a HTML image tag with a small flag for a given country code.

    Args:
        cc: Series of country codes.
        width: The desired width of the flag.

    Returns:
        Series of HTML image tags.
    """
    flags = (
        "<img "
        + (' src="' + gen_flag_url(cc, width) + '"')
        + (
            (' srcset="' + gen_flag_url(cc, width * 2) + ' 2x"')
            if width != 2560
            else ""
        )
        + f' width="{width}"'
        + (' alt="' + countries_to_scheme(cc, GeoScheme.country_name) + '"')
        + ' style="border: 1px solid #00000080"'
        + "/>"
    )

    return flags.where(~cc.isin(["not found", "", "?"]) & cc.notna(), "")



"""Types and functions for data conflict management."""

from typing import Any, Literal, TypeAlias

DataConflictPolicy: TypeAlias = Literal["raise", "ignore", "override"]
"""Policy for handling data conflicts."""

DataConflicts: TypeAlias = dict[tuple[str, str, str], tuple[Any, Any]]
"""Conflicting values indexed by their location in a table."""


class DataConflictError(ValueError):
    """Irreconsilable conflicts during import / merging of data."""

    def __init__(self, conflicts: DataConflicts) -> None:  # noqa: D107
        self.conflicts = conflicts
        super().__init__(
            f"Conflicting values: {conflicts}"
            if len(conflicts) < 5
            else f"{len(conflicts)} in table-columns "
            + str(set((k[0], k[2]) for k in conflicts.keys()))
        )



"""Helper functions for statistical evalutation of (dataframe-based) data."""

import pandas as pd


def dist_table(
    df: pd.DataFrame, id_cols: str | list[str] | None, category_cols: str | list[str]
) -> pd.Series:
    """Return a frequency table of the distribution of unique entities.

    Entities are identified by ``id_cols``. Distribution is presented over
    unique categories in ``category_cols``.

    Args:
        df: Dataframe to evaluate.
        id_cols: Columns to identify entities by.
        category_cols: Columns to evaluate distribution over.

    Returns:
        Frequency table of the distribution of unique entities.
    """
    id_cols = (
        [id_cols]
        if isinstance(id_cols, str)
        else id_cols
        if id_cols is not None
        else [n or "index" for n in df.index.names]
    )
    category_cols = [
        *([category_cols] if isinstance(category_cols, str) else category_cols)
    ]

    counts = (
        df.reset_index()[[*category_cols, *id_cols]]
        .groupby(by=category_cols, group_keys=True)
        .apply(lambda df: len(df.drop_duplicates()))
        .rename("freq")
    )

    return counts



"""Test tables module."""

import pandas as pd
from py_research.tables import ResultTable, TableColors, TableStyle, to_html


def test_result_table():
    """Test ResultTable."""
    # Create a DataFrame
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})

    # Create a ResultTable
    table = ResultTable(df)

    # Test the render method
    rendered = to_html(table.to_styled_df())
    assert isinstance(rendered, str)
    assert "A" in rendered
    assert "B" in rendered
    assert "C" in rendered

    # Test the styles attribute
    style = TableStyle(cols=["A"], css={"font-weight": "bold"})
    table.styles.append(style)
    rendered = to_html(table.to_styled_df())
    assert "font-weight: bold;" in rendered

    # Test the labels attribute
    table.labels = {"A": "Alpha", "B": "Beta", "C": "Gamma"}
    rendered = to_html(table.to_styled_df())
    assert "Alpha" in rendered
    assert "Beta" in rendered
    assert "Gamma" in rendered

    # Test widths attribute
    table.widths = {"A": "10rem", "B": "100px", "C": "auto"}
    rendered = to_html(table.to_styled_df())
    assert "width: 10rem;" in rendered
    assert "width: 100px;" in rendered
    assert "width: auto;" in rendered

    # Test relative widths attribute
    table.widths = {"A": 1, "B": 2, "C": 2}
    rendered = to_html(table.to_styled_df())
    assert "width: 20.0%;" in rendered
    assert "width: 40.0%;" in rendered
    assert "width: 40.0%;" in rendered

    # Test the title attribute
    table.title = "Test Table"
    rendered = to_html(table.to_styled_df())
    assert "Test Table" in rendered

    # Test the table_colors attribute
    table.table_colors = TableColors(header_even="#123456")
    rendered = to_html(table.to_styled_df())
    assert "background-color: #123456;" in rendered

    # Test the table_styles attribute
    table.table_styles = {"th": {"color": "#654321"}}
    rendered = to_html(table.to_styled_df())
    assert "color: #654321 !important;" in rendered


def test_result_table_multi_df():
    """Test ResultTable."""
    # Create a DataFrame
    df = pd.DataFrame(
        {
            "A": [1, 2, 3],
            "B": [4, 5, 6],
            "C": [7, 8, 9],
            "D": [10, 11, 12],
            "E": [13, 14, 15],
        }
    ).set_index(["D", "E"])

    # Create a different DataFrame
    df2 = pd.DataFrame(
        {
            "A": [10, 20, 30],
            "B": [40, 50, 60],
            "C": [70, 80, 90],
            "D": [11, 12, 13],
            "E": [14, 15, 16],
        }
    ).set_index(["D", "E"])

    # Merge the two DataFrames
    merge_df = pd.concat({"df1": df, "df2": df2}, axis="columns")

    # Create a ResultTable
    table = ResultTable(merge_df, hide_index=["E"])

    # Test the render method
    rendered = to_html(table.to_styled_df())
    assert isinstance(rendered, str)
    assert "df1" in rendered
    assert "df2" in rendered
    assert "A" in rendered
    assert "B" in rendered
    assert "C" in rendered
    assert "D" in rendered
    assert "E" not in rendered

    # Test the styles attribute with a col filter
    style = TableStyle(cols=["A"], css={"font-weight": "bold"})
    table.styles.append(style)
    rendered = to_html(table.to_styled_df())
    assert "font-weight: bold;" in rendered

    # Test the styles attribute with a row filter
    style = TableStyle(rows=(merge_df[("df1", "A")] < 3), css={"color": "blue"})
    table.styles.append(style)
    rendered = to_html(table.to_styled_df())
    assert "color: blue;" in rendered

    # Test the styles attribute with an exact col
    style = TableStyle(cols=[("df1", "A")], css={"color": "red"})
    table.styles.append(style)
    rendered = to_html(table.to_styled_df())
    assert "color: red;" in rendered

    # Test the labels attribute
    table.labels = {"A": "Alpha", "B": "Beta", "C": "Gamma"}
    rendered = to_html(table.to_styled_df())
    assert "df1" in rendered
    assert "df2" in rendered
    assert "Alpha" in rendered
    assert "Beta" in rendered
    assert "Gamma" in rendered

    # Test widths attribute
    table.widths = {"A": "10rem", "B": "100px", "C": "auto"}
    rendered = to_html(table.to_styled_df())
    assert "width: 10rem;" in rendered
    assert "width: 100px;" in rendered
    assert "width: auto;" in rendered

    # Test the title attribute
    table.title = "Test Table"
    rendered = to_html(table.to_styled_df())
    assert "Test Table" in rendered

    # Test column flatteting
    table.column_flatten_format = "{0}: {1}"
    rendered = to_html(table.to_styled_df())
    assert "df1: Alpha" in rendered
    assert "df1: Beta" in rendered
    assert "df2: Gamma" in rendered



"""Test time module."""

from datetime import datetime

import pandas as pd
from py_research.time import datetime_to_interval_series


def test_datetime_to_interval_series():
    """Test datetime_to_interval_series."""
    # Test with a Series of datetime objects
    dt_series = pd.Series(
        [
            datetime(2021, 1, 1),
            datetime(2021, 1, 2),
            datetime(2021, 1, 3),
            datetime(2022, 1, 4),
            datetime(2023, 1, 5),
        ]
    )
    interval_series = datetime_to_interval_series(dt_series)
    assert interval_series.equals(pd.Series([2021, 2021, 2021, 2022, 2023]))



"""Test ranking module."""

import pandas as pd

from py_research.ranking import create_rank_series, create_ranking_filter


def test_create_rank_series():
    """Test create_rank_series."""
    # Test with a Series
    rank_by = pd.Series([1, 2, 3, 4, 5])
    exclude = pd.Series([False, False, True, False, False])
    rank_series = create_rank_series(rank_by, exclude=exclude)
    pd.testing.assert_series_equal(
        rank_series, pd.Series([4, 3, None, 2, 1], name="rank"), check_dtype=False
    )

    # Test with a DataFrame
    rank_by = pd.DataFrame({"a": [1, 2, 3, 4, 5], "b": [5, 4, 3, 2, 1]})
    exclude = pd.Series([False, False, True, False, False])
    rank_series = create_rank_series(rank_by, exclude=exclude)
    pd.testing.assert_series_equal(
        rank_series, pd.Series([4, 3, None, 2, 1], name="rank"), check_dtype=False
    )

    # Test with ascending rank mode
    rank_series = create_rank_series(rank_by, rank_mode="ascending", exclude=exclude)
    pd.testing.assert_series_equal(
        rank_series, pd.Series([1, 2, None, 3, 4], name="rank"), check_dtype=False
    )

    # Test with no exclusion
    rank_series = create_rank_series(rank_by, rank_mode="ascending")
    pd.testing.assert_series_equal(
        rank_series, pd.Series([1, 2, 3, 4, 5], name="rank"), check_dtype=False
    )


def test_create_ranking_filter():
    """Test create_ranking_filter."""
    # Test with a Series
    rank_by = pd.Series([1, 2, 3, 4, 5])
    cutoff_rank = 3
    filter_series = create_ranking_filter(rank_by, cutoff_rank)
    pd.testing.assert_series_equal(
        filter_series, pd.Series([False, False, True, True, True]), check_names=False
    )

    # Test with ascending sort order
    filter_series = create_ranking_filter(rank_by, cutoff_rank, sort_order="ascending")
    pd.testing.assert_series_equal(
        filter_series, pd.Series([True, True, True, False, False]), check_names=False
    )

    # Test with rank_only
    rank_only = pd.Series([False, False, True, True, True])
    filter_series = create_ranking_filter(rank_by, cutoff_rank, rank_only=rank_only)
    pd.testing.assert_series_equal(
        filter_series, pd.Series([False, False, True, True, True]), check_names=False
    )

    # Test with show_only
    show_only = pd.Series([False, False, True, True, True])
    filter_series = create_ranking_filter(rank_by, cutoff_rank, show_only=show_only)
    pd.testing.assert_series_equal(
        filter_series, pd.Series([False, False, True, True, True]), check_names=False
    )

    # Test with show_always
    show_always = pd.Series([True, False, False, False, False])
    filter_series = create_ranking_filter(rank_by, cutoff_rank, show_always=show_always)
    pd.testing.assert_series_equal(
        filter_series, pd.Series([True, False, True, True, True]), check_names=False
    )

    # Test with count_always_shown
    filter_series = create_ranking_filter(
        rank_by, cutoff_rank, show_always=show_always, count_always_shown=True
    )
    pd.testing.assert_series_equal(
        filter_series, pd.Series([True, False, True, True, True]), check_names=False
    )



"""Utilities for importing different data representations into relational format."""

from collections.abc import Callable, Hashable, Mapping
from dataclasses import dataclass
from datetime import datetime
from itertools import chain, groupby
from typing import Any, Literal, TypeAlias, cast, overload
from uuid import uuid4

import pandas as pd

from py_research.hashing import gen_str_hash

from .base import DB
from .conflicts import DataConflictError, DataConflictPolicy, DataConflicts

Scalar = str | int | float | datetime

_AttrMap = Mapping[str, "str | bool | _AttrMap"]
"""Mapping of hierarchical attributes to table columns"""

_RelationalMap = Mapping[str, "str | bool | _RelationalMap | TableMap | list[TableMap]"]
"""Mapping of hierarchical attributes to table columns or other tables."""


@dataclass
class TableMap:
    """Configuration for how to map (nested) dictionary items to relational tables."""

    table: str
    """Name of the table to map to attributes in ``map`` to."""

    map: _RelationalMap | set[str] | str | Callable[
        [dict | str], _RelationalMap | set[str] | str
    ]
    """Mapping of hierarchical attributes to table columns or other tables."""

    ext_maps: "list[TableMap] | None" = None
    """Map attributes on the same level to different tables"""

    link_attr: str | None = None
    """Override attribute to use when referencing this table from a parent table."""

    link_type: Literal["1-n", "n-1", "n-m"] = "n-m"
    """Type of reference between this table and the parent table.
    Only ``n-m`` will create a join table.
    """

    join_table_name: str | None = None
    """Name of the join table to use for referencing this table to its parent, if any"""

    join_table_map: _AttrMap | None = None
    """Mapping of hierarchical attributes to join table rows."""

    id_type: Literal["hash", "attr", "uuid"] = "hash"
    """Type of id to use for this table.
    - ``hash``: sha256 hash of a subset of all mapped attributes
    - ``attr``: use given attr as id directly, no hashing
    - ``uuid``: generate random uuid for each row
    """

    id_attr: str | list[str] | None = None
    """Name of unique, mapped attribute to use as id directly or
    list of mapped attributes to use for auto-generating row ids via sha256 hashing.
    If None, all mapped attributes will be used for hashing.
    """

    hash_id_with_path: bool = False
    """If True, the id will be generated based on the data and the full tree path.
    """

    match_by_attr: bool | str = False
    """Try to match this mapped data to target table (by given attr)
    before creating a new row.
    """

    conflict_policy: DataConflictPolicy = "raise"
    """Which policy to use if import conflicts occur for this table."""


_SubMap = tuple[dict | list, TableMap | list[TableMap]]
"""Combination of data and mapping for a subtree."""


def _map_sublayer(
    node: dict, mapping: _RelationalMap
) -> tuple[pd.Series, dict[str, _SubMap]]:
    """Extract hierarchical data into set of scalar attributes + ref'd data objects."""
    # Split the current mapping level into groups based on type.
    target_groups: dict[type, dict] = {
        t: dict(g)  # type: ignore
        for t, g in groupby(
            sorted(
                mapping.items(),
                key=lambda item: str(type(item[1])),
            ),
            key=lambda item: type(item[1]),
        )
    }  # type: ignore

    # First list and handle all scalars, hence data attributes on the current level,
    # which are to be mapped to table columns.
    scalars = dict(
        chain(
            (target_groups.get(str) or {}).items(),
            (target_groups.get(bool) or {}).items(),
        )
    )

    cols = {
        (col if isinstance(col, str) else attr): data
        for attr, col in scalars.items()
        if isinstance(data := node.get(attr), Scalar)
    }

    refs = {
        attr: (data, cast(TableMap | list[TableMap], sub_map))
        for attr, sub_map in {
            **(target_groups.get(TableMap) or {}),
            **(target_groups.get(list) or {}),
        }.items()
        if isinstance(data := node.get(attr), dict | list)
    }

    # Handle nested data attributes (which come as dict types).
    for attr, sub_map in (target_groups.get(dict) or {}).items():
        sub_node = node.get(attr)
        if isinstance(sub_node, dict):
            sub_row, sub_refs = _map_sublayer(sub_node, cast(dict, sub_map))
            cols = {**cols, **sub_row}
            refs = {**refs, **sub_refs}

    return pd.Series(cols, dtype=object), refs


def _gen_row_hash(
    row: pd.Series,
    context_path: list[str | int] | None = None,
    hash_subset: list[str] | None = None,
) -> str:
    """Generate hash for given row."""
    hash_row = (
        row[list(set(hash_subset) & set(row.index))] if hash_subset is not None else row
    )
    row_id = gen_str_hash(
        hash_row.to_dict()
        if context_path is not None
        else (hash_row.to_dict(), context_path)
    )

    return row_id


_DictDB: TypeAlias = dict[str, dict[Hashable, dict[str, Any]]]
"""All-dictionary representation of a relational database for performant extending."""

_Rels: TypeAlias = dict[tuple[str, str], tuple[str, str]]
"""Relations between tables, indexed by the referencing table and column."""

_JoinTables: TypeAlias = set[str]
"""Names of all join tables in the database."""

RelDB = tuple[_DictDB, _Rels, _JoinTables]
"""Full relational database representation."""


def _handle_refs(  # noqa: C901
    mapping: TableMap,
    db: RelDB,
    row: pd.Series,
    refs: list[tuple[str | None, _SubMap]],
    collect_conflicts: bool = False,
    _all_conflicts: DataConflicts | None = None,
    _path: list[str | int] | None = None,
) -> DataConflicts:
    """Handle references to other tables."""
    _all_conflicts = _all_conflicts or {}
    _path = _path or []
    database, relations, join_tables = db

    # Handle nested data, which is to be extracted into separate tables and referenced.
    for attr, (sub_data, sub_maps) in refs:
        # Get info about the ref table to use from mapping
        # (or generate a new for the ref table).

        if not isinstance(sub_maps, list):
            sub_maps = [sub_maps]

        for sub_map in sub_maps:
            join_table_name = sub_map.join_table_name
            join_table_exists = False
            alt_join_table_names = [
                f"{mapping.table}_{sub_map.table}",
                f"{sub_map.table}_{mapping.table}",
            ]

            if join_table_name is None:
                join_table_name = alt_join_table_names[0]
                for name in alt_join_table_names:
                    if name in database:
                        join_table_name = name
                        join_table_exists = True
                        break
            else:
                join_table_exists = join_table_name in database

            if not isinstance(sub_data, list):
                sub_data = [sub_data]

            for i, sub_data_item in enumerate(sub_data):
                _sub_path = [_path, attr, i]

                rel_row, _all_conflicts = _tree_to_db(
                    sub_data_item,
                    sub_map,
                    db,
                    collect_conflicts,
                    _all_conflicts,
                    _sub_path,
                )

                if sub_map.link_type == "n-m":
                    # Map via join table.
                    if not join_table_exists:
                        database[join_table_name] = {}
                        join_table_exists = True

                    ref_row, _ = (
                        _map_sublayer(sub_data_item, sub_map.join_table_map)
                        if isinstance(sub_data_item, dict)
                        and sub_map.join_table_map is not None
                        else (pd.Series(dtype=object), None)
                    )

                    left_col = f"{attr}_of" if attr is not None else mapping.table
                    relations[(join_table_name, left_col)] = (
                        mapping.table,
                        "_id",
                    )
                    ref_row[left_col] = row.name

                    right_col = attr if attr is not None else sub_map.table
                    relations[(join_table_name, right_col)] = (
                        sub_map.table,
                        "_id",
                    )
                    ref_row[right_col] = rel_row.name

                    ref_row.name = _gen_row_hash(ref_row, _sub_path)

                    database[join_table_name][ref_row.name] = ref_row.to_dict()
                    join_tables |= {join_table_name}
                elif sub_map.link_type == "1-n":
                    # Map via direct reference from children to parent.
                    col = f"{attr}_of" if attr is not None else mapping.table

                    if col in rel_row.index:
                        assert rel_row[col] is None or rel_row[col] == row.name

                    relations[(sub_map.table, col)] = (
                        mapping.table,
                        "_id",
                    )

                    database[sub_map.table][rel_row.name] = {
                        **rel_row.to_dict(),
                        col: row.name,
                    }
                elif sub_map.link_type == "n-1":
                    # Map via direct reference from parents to child.
                    col = attr if attr is not None else sub_map.table

                    if col in row.index:
                        assert row[col] is None or row[col] == row.name

                    relations[(mapping.table, col)] = (
                        sub_map.table,
                        "_id",
                    )

                    database[mapping.table][row.name] = {
                        **row.to_dict(),
                        col: rel_row.name,
                    }

    return _all_conflicts


def _tree_to_db(  # noqa: C901
    data: dict | str,
    mapping: TableMap,
    db: RelDB,
    collect_conflicts: bool = False,
    _all_conflicts: DataConflicts | None = None,
    _path: list[str | int] | None = None,
) -> tuple[pd.Series, DataConflicts]:
    """Transform recursive dictionary data into relational format."""
    _all_conflicts = _all_conflicts or {}
    _path = _path or []
    database, _, _ = db

    # Initialize new table as defined by mapping, if it doesn't exist yet.
    if mapping.table not in database:
        database[mapping.table] = {}

    resolved_map = (
        mapping.map(data) if isinstance(mapping.map, Callable) else mapping.map
    )

    # Extract row of data attributes and refs to other objects.
    row = None
    refs: list[tuple[str | None, _SubMap]] = []
    # If mapping is only a string, extract the target attr directly.
    if isinstance(data, str):
        assert isinstance(resolved_map, str)
        row = pd.Series({resolved_map: data}, dtype=object)
    else:
        if isinstance(resolved_map, set):
            row = pd.Series(
                {k: v for k, v in data.items() if k in resolved_map}, dtype=object
            )
        elif isinstance(resolved_map, dict):
            row, ref_dict = _map_sublayer(data, resolved_map)
            refs = [*refs, *ref_dict.items()]
        else:
            raise TypeError(
                f"Unsupported mapping type {type(resolved_map)}"
                f" for data of type {type(data)}"
            )

    row.name = (
        row[mapping.id_attr]
        if mapping.id_type == "attr" and isinstance(mapping.id_attr, str)
        else _gen_row_hash(
            row,
            _path,
            [mapping.id_attr] if isinstance(mapping.id_attr, str) else mapping.id_attr,
        )
        if mapping.id_type == "hash"
        else str(uuid4())[-10:]
    )

    if not isinstance(row.name, str | int):
        raise ValueError(
            f"Value of `'{mapping.id_attr}'` (`TableMap.id_attr`) "
            f"must be a string or int for all objects, but received {row.name}"
        )

    if mapping.ext_maps is not None:
        assert isinstance(data, dict)
        refs = [*refs, *((m.link_attr, (data, m)) for m in mapping.ext_maps)]

    _all_conflicts = _handle_refs(
        mapping, db, row, refs, collect_conflicts, _all_conflicts
    )

    if not row.empty:
        existing_row: dict[str, Any] | None = None
        if mapping.match_by_attr:
            # Make sure any existing data in database is consistent with new data.
            match_by = cast(str, mapping.match_by_attr)
            if mapping.match_by_attr is True:
                assert isinstance(resolved_map, str)
                match_by = resolved_map

            match_to = row[match_by]

            # Match to existing row or create new one.
            existing_rows: list[tuple[str, dict[str, Any]]] = list(
                filter(
                    lambda i: i[1][match_by] == match_to,
                    database[mapping.table].items(),
                )
            )

            if len(existing_rows) > 0:
                existing_row_id, existing_row = existing_rows[0]
                # Take over the id of the existing row.
                row.name = existing_row_id
        else:
            existing_row = database[mapping.table].get(row.name)

        if existing_row is not None:
            existing_attrs = set(
                str(k) for k, v in existing_row.items() if k and pd.notna(v)
            )
            new_attrs = set(str(k) for k, v in row.items() if k and pd.notna(v))

            # Assert that overlapping attributes are equal.
            intersect = existing_attrs & new_attrs

            if mapping.conflict_policy == "raise":
                conflicts = {
                    (mapping.table, row.name, c): (existing_row[c], row[c])
                    for c in intersect
                    if existing_row[c] != row[c]
                }

                if len(conflicts) > 0:
                    if not collect_conflicts:
                        raise DataConflictError(conflicts)

                    _all_conflicts = {**_all_conflicts, **conflicts}
            if mapping.conflict_policy == "ignore":
                row = pd.Series(
                    {**row.loc[list(new_attrs)], **existing_row}, name=row.name
                )
            else:
                row = pd.Series(
                    {**existing_row, **row.loc[list(new_attrs)]}, name=row.name
                )

        # Add row to database table or update it.
        database[mapping.table][row.name] = row.to_dict()

    # Return row (used for recursion).
    return row, _all_conflicts


@overload
def tree_to_db(
    data: dict | str,
    mapping: TableMap,
    collect_conflicts: Literal[True] = ...,
) -> tuple[DB, DataConflicts]:
    ...


@overload
def tree_to_db(
    data: dict | str,
    mapping: TableMap,
    collect_conflicts: Literal[False] = ...,
) -> DB:
    ...


def tree_to_db(  # noqa: C901
    data: dict | str,
    mapping: TableMap,
    collect_conflicts: bool = False,
) -> DB | tuple[DB, DataConflicts]:
    """Transform recursive dictionary data into relational format.

    Args:
        data: The data to be transformed
        mapping:
            Configuration for how to performm the mapping.
        collect_conflicts:
            Collect all conflicts and return them, rather than stopping right away.

    Returns:
        The relational database representation of the data.
        If ``collect_conflicts`` is ``True``, a tuple of the database and the conflicts
        is returned.
    """
    df_dict = {}
    rels = {}
    join_tables = set()
    _, conflicts = _tree_to_db(
        data, mapping, (df_dict, rels, join_tables), collect_conflicts
    )
    db = DB(
        table_dfs={
            name: pd.DataFrame.from_dict(df, orient="index").rename_axis(
                "_id", axis="index"
            )
            for name, df in df_dict.items()
        },
        relations=rels,
        join_tables=join_tables,
    )
    return (db, conflicts) if collect_conflicts else db



"""Test hashing module."""

from datetime import date
from decimal import Decimal
from fractions import Fraction

from py_research.hashing import gen_str_hash


def test_gen_str_hash():
    """Test gen_str_hash."""
    # Test with different types of numbers
    assert len(gen_str_hash(123, length=10)) == 10
    assert len(gen_str_hash(123.456, length=10)) == 10
    assert len(gen_str_hash(complex(1, 2), length=10)) == 10
    assert len(gen_str_hash(Decimal("123.456"), length=10)) == 10
    assert len(gen_str_hash(Fraction(1, 2), length=10)) == 10

    # Test with a string
    assert len(gen_str_hash("test", length=10)) == 10

    # Test with a date
    assert len(gen_str_hash(date.today(), length=10)) == 10



"""Test intl module."""

from datetime import datetime
from typing import Any

import pandas as pd
import pytest
from py_research.intl import Args, DtUnit, Format, Overrides, iter_locales


@pytest.fixture
def locales() -> list[str]:
    """Return sample locales."""
    return ["de_DE", "en_US"]


@pytest.fixture
def overrides() -> dict[str, Overrides]:
    """Return sample translation overrides."""
    return {
        "en_US": Overrides(
            {
                "car": "automobile",
                "house": "mouse",
            },
            templates={
                "vehicles": {("bike", "car"): "Vehicle: {0}"},
                "buildings": {Args({"house", "bridge"}): "Building: {0}"},
            },
        ),
        "de_DE": Overrides({"car": "Karren", "house": "Haus"}),
    }


@pytest.mark.parametrize(
    "locale, ctx, label, expected",
    [
        ("en_US", "vehicles", "car", "Vehicle: automobile"),
        ("en_US", None, "bike", "bike"),
        ("en_US", None, "house", "mouse"),
        ("de_DE", "vehicles", "car", "Fahrzeug: Karren"),
        ("de_DE", None, "car", "Karren"),
        ("de_DE", "vehicles", "bike", "Fahrzeug: Fahrrad"),
        ("de_DE", None, "bike", "Fahrrad"),
        ("de_DE", None, "house", "Haus"),
        ("de_DE", None, "box", "Kasten"),
        ("de_DE", "buildings", "house", "Gebäude: Haus"),
    ],
)
def test_localize_label(
    locales: list[str],
    overrides: dict[str, Overrides],
    locale: str,
    ctx: str | None,
    label: str,
    expected: str,
):
    """Test localization of text."""
    for loc in iter_locales(locales, overrides):
        if str(loc.locale) == locale:
            assert loc.label(label, context=ctx) == expected


@pytest.mark.parametrize(
    "locale, format, value, expected",
    [
        ("en_US", Format(), 1, "1"),
        ("en_US", Format(), 1.5, "1.5"),
        ("en_US", Format(decimal_min_digits=2), 1.5, "1.50"),
        ("en_US", Format(decimal_max_digits=3), 1.56789, "1.568"),
        ("en_US", Format(datetime_format=DtUnit.year), datetime(2024, 1, 23), "2024"),
        ("en_US", Format(datetime_format=DtUnit.month), datetime(2024, 1, 23), "01"),
        ("en_US", Format(datetime_format=DtUnit.day), datetime(2024, 1, 23), "23"),
        ("en_US", Format(datetime_format="yyyy-MM"), datetime(2024, 1, 23), "2024-01"),
        ("en_US", Format(), "Hello", "Hello"),
        ("de_DE", Format(), 1, "1"),
        ("de_DE", Format(decimal_min_digits=2), 1.5, "1,50"),
        ("de_DE", Format(decimal_max_digits=3), 1.56789, "1,568"),
        ("de_DE", Format(), "Hello", "Hallo"),
        ("de_DE", Format(), None, ""),
        ("de_DE", Format(na_representation="N/A"), None, "N/A"),
        ("de_DE", Format(), pd.NA, ""),
        ("de_DE", Format(na_representation="N/A"), pd.NA, "N/A"),
    ],
)
def test_localize_value(
    locales: list[str],
    locale: str,
    format: Format,
    value: Any,
    expected: str,
):
    """Test localization of values."""
    for loc in iter_locales(locales):
        if str(loc.locale) == locale:
            assert loc.value(value, format) == expected


def test_auto_digits():
    """Test auto digits."""
    f = Format().auto_digits([1.5, 1.56789, 1.56789, 1.56789, 1.56789])
    assert f.decimal_min_digits == 2
    assert f.decimal_max_digits == 2

    f2 = Format().auto_digits([1.5, 1.56789, 1.56789, 1.56789, 1.56789], fixed=False)
    assert f2.decimal_min_digits == 2
    assert f2.decimal_max_digits is None
